"""Hybrid programs"""

from collections import OrderedDict
from typing import Union

from ss2hcsp.hcsp.expr import Expr, AVar, AConst, Expr, true_expr, false_expr, RelExpr, LogicExpr, Expr
from ss2hcsp.hcsp import assertion
from ss2hcsp.util.topsort import topological_sort
import re


class Channel:
    """Models a channel identifier. It is a string followed by a list
    of integer or variable arguments.

    The usual channel is modeled by a string without arguments. Further
    arguments can be given for parameterized channels. Parameters that
    are already decided are indicated by integers. Those that still need
    to be matched are indicated by variables.

    """
    def __init__(self, name, args=None, meta=None):
        assert isinstance(name, str)
        if args is None:
            args = tuple()

        # Each argument is either integer, (unevaluated) expression, or variable
        assert isinstance(args, tuple) and all(isinstance(arg, (Expr, int, str)) for arg in args)

        self.name = name
        self.args = args
        self.meta = meta

    def __str__(self):
        def str_arg(arg):
            # if isinstance(arg, str):
            #     return repr(arg)
            if isinstance(arg, AConst) and isinstance(arg.value, str) and arg.value[0] != "\"":
                return "\"" + arg.value + "\""
            else:
                return str(arg)
        return self.name + ''.join('[' + str_arg(arg) + ']' for arg in self.args)
    
    def __repr__(self):
        if self.args:
            return "Channel(%s,%s)" % (self.name, ','.join(str(arg) for arg in self.args))
        else:
            return "Channel(%s)" % self.name

    def __hash__(self):
        return hash(("CHANNEL", self.name, self.args))

    def __eq__(self, other):
        return self.name == other.name and self.args == other.args

    def __le__(self, other):
        return (self.name, self.args) <= (other.name, other.args)

    def __lt__(self, other):
        return (self.name, self.args) < (other.name, other.args)

    def subst(self, inst):
        if self.name in inst:
            target = inst[self.name]
            assert isinstance(target, Channel)
            return Channel(target.name, tuple(arg if isinstance(arg, (int, str)) else arg.subst(inst)
                                              for arg in target.args + self.args))
        else:
            return Channel(self.name, tuple(arg if isinstance(arg, (int, str)) else arg.subst(inst)
                                            for arg in self.args))

class Function:
    """Declarations of (pure functions).
    
    Each pure functions is defined by a list of variables and an expression.
    When the function is called, the variables in the expression are
    substituted for the arguments to the function.
    
    """
    def __init__(self, name: str, vars, expr: Union[Expr, str]):
        assert isinstance(vars, list) and all(isinstance(var, str) for var in vars)
        if isinstance(expr, str):
            from ss2hcsp.hcsp.parser import expr_parser
            expr = expr_parser.parse(expr)
        self.name = name
        self.vars = vars
        self.expr = expr

    def __str__(self):
        return "%s(%s) = %s" % (self.name, ",".join(var for var in self.vars), self.expr)

    def __repr__(self):
        return "Function(%s, %s, %s)" % (self.name, repr(self.vars), repr(self.expr))


class HCSP:
    def __init__(self):
        self.type = None

    def sc_str(self):
        """
        The HCSP form that can be accepted by HCSP2SC,
        the program translating HCSP into SystemC.
        """
        return str(self)

    def get_vars(self):
        return set()

    def get_fun_names(self):
        return set()

    def get_zero_arity_funcs(self):
        return set()

    def get_chs(self):
        return set()

    def priority(self):
        if self.type == "parallel":
            return 30
        elif self.type == "select_comm":
            return 50
        elif self.type == "sequence":
            return 70
        elif self.type == "ichoice":
            return 80
        else:
            return 100

    def size(self):
        """Returns size of HCSP program."""
        if isinstance(self, (Var, Skip, Wait, Assign, Assert, Test, Log,
                             InputChannel, OutputChannel)):
            return 1
        elif isinstance(self, (Sequence, Parallel)):
            return 1 + sum([hp.size() for hp in self.hps])
        elif isinstance(self, Loop):
            return 1 + self.hp.size()
        elif isinstance(self, ODE):
            return 1
        elif isinstance(self, (ODE_Comm, SelectComm)):
            return 1 + sum([1 + out_hp.size() for comm_hp, out_hp in self.io_comms])
        elif isinstance(self, ITE):
            return 1 + sum([1 + hp.size() for cond, hp in self.if_hps]) + \
              (self.else_hp.size() if self.else_hp else 0)
        else:
            raise NotImplementedError

    def contain_hp(self, name):
        """Returns whether the given HCSP program contains Var(name)."""
        if isinstance(self, Var):
            return self.name == name
        elif isinstance(self, (Skip, Wait, Assign, Assert, Test, Log, InputChannel, OutputChannel)):
            return False
        elif isinstance(self, (Sequence, Parallel)):
            for sub_hp in self.hps:
                if sub_hp.contain_hp(name):
                    return True
        elif isinstance(self, Loop):
            return self.hp.contain_hp(name)
        elif isinstance(self, ODE):
            return self.out_hp.contain_hp(name)
        elif isinstance(self, (ODE_Comm, SelectComm)):
            for io_comm in self.io_comms:
                if io_comm[1].contain_hp(name):
                    return True
        elif isinstance(self, ITE):
            if self.else_hp.else_hp is not None and self.else_hp.contain_hp(name):
                return True
            for sub_hp in [hp for cond, hp in self.if_hps]:
                if sub_hp.contain_hp(name):
                    return True
        else:
            raise NotImplementedError

    def get_contain_hps_count(self):
        """Returns the dictionary mapping contained hps to number
        of appearances.
        
        """
        res = dict()

        def rec(hp):
            if isinstance(hp, Var):
                if hp.name not in res:
                    res[hp.name] = 0
                res[hp.name] += 1
            elif isinstance(hp, (Skip, Wait, Assign, Assert, Test, Log, InputChannel, OutputChannel)):
                pass
            elif isinstance(hp, (Sequence, Parallel)):
                for sub_hp in hp.hps:
                    rec(sub_hp)
            elif isinstance(hp, Loop):
                rec(hp.hp)
            elif isinstance(hp, ODE):
                rec(hp.out_hp)
            elif isinstance(hp, (ODE_Comm, SelectComm)):
                for _, out_hp in hp.io_comms:
                    rec(out_hp)
            elif isinstance(hp, ITE):
                for _, sub_hp in hp.if_hps:
                    rec(sub_hp)
                if hp.else_hp is not None:
                    rec(hp.else_hp)
            else:
                raise NotImplementedError
        rec(self)
        return res

    def get_contain_hps(self):
        """Returns the set of Var calls contained in self."""
        return set(self.get_contain_hps_count())

    def subst_comm(self, inst):
        def subst_io_comm(io_comm):
            return (io_comm[0].subst_comm(inst), io_comm[1].subst_comm(inst))

        def subst_if_hp(if_hp):
            return (if_hp[0].subst(inst), if_hp[1].subst_comm(inst))

        def subst_ode_eq(ode_eq):
            return (ode_eq[0], ode_eq[1].subst(inst))

        if self.type in ('var', 'skip'):
            return self
        elif self.type == 'wait':
            return Wait(self.delay.subst(inst))
        elif self.type == 'assign':
            return Assign(self.var_name, self.expr.subst(inst))
        elif self.type == 'assert':
            return Assert(self.bexpr.subst(inst), [expr.subst(inst) for expr in self.msgs])
        elif self.type == 'test':
            return Test(self.bexpr.subst(inst), [expr.subst(inst) for expr in self.msgs])
        elif self.type == 'log':
            return Log(self.pattern.subst(inst), exprs=[expr.subst(inst) for expr in self.exprs])
        elif self.type == 'input_channel':
            return InputChannel(self.ch_name.subst(inst), self.var_name)
        elif self.type == 'output_channel':
            if self.expr is None:
                return OutputChannel(self.ch_name.subst(inst))
            else:
                return OutputChannel(self.ch_name.subst(inst), self.expr.subst(inst))
        elif self.type == 'sequence':
            return Sequence(*(hp.subst_comm(inst) for hp in self.hps))
        elif self.type == 'ode':
            return ODE([subst_ode_eq(eq) for eq in self.eqs],
                       self.constraint.subst(inst), out_hp=self.out_hp.subst_comm(inst))
        elif self.type == 'ode_comm':
            return ODE_Comm([subst_ode_eq(eq) for eq in self.eqs],
                            self.constraint.subst(inst),
                            [subst_io_comm(io_comm) for io_comm in self.io_comms])
        elif self.type == 'loop':
            return Loop(self.hp.subst_comm(inst), constraint=self.constraint)
        elif self.type == 'select_comm':
            return SelectComm(*(subst_io_comm(io_comm) for io_comm in self.io_comms))
        elif self.type == 'ite':
            return ITE([subst_if_hp(if_hp) for if_hp in self.if_hps], 
              self.else_hp.subst_comm(inst) if self.else_hp is not None else None)
        else:
            print(self.type)
            raise NotImplementedError


class Var(HCSP):
    def __init__(self, name, meta=None):
        super(Var, self).__init__()
        self.type = "var" 
        assert isinstance(name, str)
        self.name = name
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.name == other.name

    def __repr__(self):
        return "Var(%s)" % self.name

    def __str__(self):
        return "@" + self.name + ";"

    def __hash__(self):
        return hash(("VAR", self.name))

    def sc_str(self):
        return self.name


class Skip(HCSP):
    def __init__(self, meta=None):
        super(Skip, self).__init__()
        self.type = "skip"
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type

    def __repr__(self):
        return "Skip()"

    def __str__(self):
        return "skip;"

    def __hash__(self):
        return hash("Skip")


class Wait(HCSP):
    def __init__(self, delay, meta=None):
        super(Wait, self).__init__()
        assert isinstance(delay, Expr)
        self.type = "wait"
        self.delay = delay
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.delay == other.delay

    def __repr__(self):
        return "Wait(%s)" % str(self.delay)

    def __str__(self):
        return "wait(%s);" % str(self.delay)

    def __hash__(self):
        return hash(("Wait", self.delay))


class Assign(HCSP):
    """Assignment command.

    var_name : Expr - variable to be assigned, one of AVar (e.g. x),
        ArrayIdxExpr (e.g. a[0]), and FieldNameExpr (e.g. a.field1).
    expr : Expr - value to be assigned.

    Left side is an expression that can serve as a lname. This includes
    variables, array indices, and field names.

    """
    def __init__(self, var_name, expr, meta=None):
        super(Assign, self).__init__()
        self.type = "assign"
        assert isinstance(expr, (Expr, str, int))
        if isinstance(var_name, str):
            var_name = AVar(str(var_name))
        if isinstance(var_name, Expr):
            self.var_name = var_name
        else:
            var_name = tuple(var_name)
            assert len(var_name) >= 2 and all(isinstance(name, (str, Expr)) for name in var_name)
            self.var_name = [AVar(n) if isinstance(n, str) else n for n in var_name]
        self.expr = expr
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.var_name == other.var_name and self.expr == other.expr

    def __repr__(self):
        if isinstance(self.var_name, Expr):
            var_str = str(self.var_name)
        else:
            var_str = "[%s]" % (','.join(str(n) for n in self.var_name))
        return  "Assign(%s,%s)" % (var_str, str(self.expr))

    def __str__(self):
        if isinstance(self.var_name, Expr):
            var_str = str(self.var_name)
        else:
            var_str = "(%s)" % (', '.join(str(n) for n in self.var_name))
        return "%s := %s;" % (var_str, self.expr)

    def __hash__(self):
        return hash(("Assign", self.var_name, self.expr))

    def get_vars(self):
        if isinstance(self.var_name, Expr):
            var_set = {str(self.var_name)}
        else:
            var_set = set(str(n) for n in self.var_name)
        return var_set.union(self.expr.get_vars())

    def get_fun_names(self):
        return set().union(self.expr.get_fun_names())

    def get_zero_arity_funcs(self):
        return set().union(self.expr.get_zero_arity_funcs())

    def sc_str(self):
        return re.sub(pattern=":=", repl="=", string=str(self))

    def priority(self):
        return 100

class RandomAssign(HCSP):
    """Random Assignment commend.

    x := {x >= 1}

    var_name : Expr - variable to be assigned, one of AVar (e.g. x), ArrayIdxExpr (e.g. a[0]), and FieldNameExpr (e.g. a.field1).
    expr : Expr - the range value to be randomly choose from.

    Left side is an expression that can serve as a lname. This includes
    variables, array indices, and field names.

    """
    def __init__(self, var_name, expr, meta=None):
        super(RandomAssign, self).__init__()
        self.type = "randomassign"
        assert isinstance(expr, Expr)
        if isinstance(var_name, str):
            var_name = AVar(str(var_name))
        if isinstance(var_name, Expr):
            self.var_name = var_name
        else:
            var_name = tuple(var_name)
            assert len(var_name) <= 2 and all (isinstance(name, (str, Expr))for name in var_name)
            self.var_name = [AVar(n) if isinstance(n, str) else n for n in var_name]
        self.expr = expr
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.var_name == other.var_name and self.expr == other.expr

    def __repr__(self):
        if isinstance(self.var_name, Expr):
            var_str = str(self.var_name)
        else:
            var_str = "[%s]" % (','.join(str(n) for n in self.var_name))
        return "RandomAssign(%s, %s)" % (var_str, str(self.expr))

    def __str__(self):
        if isinstance(self.var_name, Expr):
            var_str = str(self.var_name)
        else:
            var_str = "(%s)" % (', '.join(str(n) for n in self.var_name))
        return var_str + " := " + "*(" + str(self.expr) + ");"

    def __hash__(self):
        return hash(("RandomAssign", self.var_name, self.expr))

    def get_vars(self):
        if isinstance(self.var_name, Expr):
            var_set = {str(self.var_name)}
        else:
            var_set = set(str(n) for n in self.var_name)
        return var_set.union(self.expr.get_vars())

    def get_fun_names(self):
        return set.union(self.expr.get_fun_names())

    def get_zero_arity_funcs(self):
        return set.union(self.expr.get_zero_arity_funcs())

    def priority(self):
        return 100

class Assert(HCSP):
    def __init__(self, bexpr, msgs, meta=None):
        super(Assert, self).__init__()
        self.type = "assert"
        assert isinstance(bexpr, Expr)
        self.bexpr = bexpr
        msgs = tuple(msgs)
        assert all(isinstance(msg, Expr) for msg in msgs)
        self.msgs = msgs
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.bexpr == other.bexpr and self.msgs == other.msgs

    def __repr__(self):
        if self.msgs:
            return "Assert(%s,%s)" % (repr(self.bexpr), ','.join(repr(msg) for msg in self.msgs))
        else:
            return "Assert(%s)" % repr(self.bexpr)

    def __str__(self):
        if self.msgs:
            return "assert(%s,%s);" % (self.bexpr, ','.join(str(msg) for msg in self.msgs))
        else:
            return "assert(%s);" % self.bexpr

    def __hash__(self):
        return hash(("Assert", self.bexpr, self.msgs))

    def get_vars(self):
        var_set = self.bexpr.get_vars()
        for msg in self.msgs:
            var_set.update(msg.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = self.bexpr.get_fun_names()
        for msg in self.msgs:
            fun_set.update(msg.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = self.bexpr.get_zero_arity_funcs()
        for msg in self.msgs:
            fun_set.update(msg.get_zero_arity_funcs())
        return fun_set


class Test(HCSP):
    def __init__(self, bexpr, msgs, meta=None):
        super(Test, self).__init__()
        self.type = "test"
        assert isinstance(bexpr, Expr)
        self.bexpr = bexpr
        msgs = tuple(msgs)
        assert all(isinstance(msg, Expr) for msg in msgs)
        self.msgs = msgs
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.bexpr == other.bexpr and self.msgs == other.msgs

    def __repr__(self):
        if self.msgs:
            return "Test(%s,%s)" % (repr(self.bexpr), ','.join(repr(msg) for msg in self.msgs))
        else:
            return "Test(%s)" % repr(self.bexpr)

    def __str__(self):
        if self.msgs:
            return "test(%s,%s);" % (self.bexpr, ','.join(str(msg) for msg in self.msgs))
        else:
            return "test(%s);" % self.bexpr

    def __hash__(self):
        return hash(("Test", self.bexpr, self.msgs))

    def get_vars(self):
        var_set = self.bexpr.get_vars()
        for msg in self.msgs:
            var_set.update(msg.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = self.bexpr.get_fun_names()
        for msg in self.msgs:
            fun_set.update(msg.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = self.bexpr.get_zero_arity_funcs()
        for msg in self.msgs:
            fun_set.update(msg.get_zero_arity_funcs())
        return fun_set


class Log(HCSP):
    def __init__(self, pattern, *, exprs=None, meta=None):
        super(Log, self).__init__()
        self.type = "log"
        assert isinstance(pattern, Expr)
        self.pattern = pattern
        if exprs is None:
            exprs = tuple()
        else:
            exprs = tuple(exprs)
        assert all(isinstance(expr, Expr) for expr in exprs)
        self.exprs = exprs
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.pattern == other.pattern and self.exprs == other.exprs

    def __repr__(self):
        if self.exprs:
            return "Log(%s,%s)" % (repr(self.pattern), ','.join(repr(expr) for expr in self.exprs))
        else:
            return "Log(%s)" % repr(self.pattern)

    def __str__(self):
        if self.exprs:
            return "log(%s,%s);" % (self.pattern, ','.join(str(expr) for expr in self.exprs))
        else:
            return "log(%s);" % self.pattern

    def __hash__(self):
        return hash(("Log", self.pattern, self.exprs))

    def get_vars(self):
        var_set = self.pattern.get_vars()
        for expr in self.exprs:
            var_set.update(expr.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = self.pattern.get_fun_names()
        for expr in self.exprs:
            fun_set.update(expr.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = self.pattern.get_zero_arity_funcs()
        for expr in self.exprs:
            fun_set.update(expr.get_zero_arity_funcs())
        return fun_set
    
class InputChannel(HCSP):
    def __init__(self, ch_name, var_name=None, meta=None):
        super(InputChannel, self).__init__()
        self.type = "input_channel"
        if isinstance(ch_name, str):
            ch_name = Channel(ch_name)
        assert isinstance(ch_name, Channel)
        if isinstance(var_name, str):
            var_name = AVar(str(var_name))
        assert var_name is None or isinstance(var_name, Expr)
        self.ch_name = ch_name  # Channel
        self.var_name = var_name  # Expr or None
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.ch_name == other.ch_name and \
            self.var_name == other.var_name

    def __repr__(self):
        if self.var_name:
            return "InputC(%s,%s)" % (self.ch_name, self.var_name)
        else:
            return "InputC(%s)" % self.ch_name

    def __str__(self):
        if self.var_name:
            return str(self.ch_name) + "?" + str(self.var_name) + ";"
        else:
            return str(self.ch_name) + "?" + ";"

    def __hash__(self):
        return hash(("InputChannel", self.ch_name, self.var_name))

    def get_vars(self):
        if self.var_name:
            return {str(self.var_name)}
        return set()

    def get_fun_names(self):
        return set()

    def get_zero_arity_funcs(self):
        return set()

    def get_chs(self):
        return {self.ch_name}

    def sc_str(self):
        return re.sub(pattern="\\?", repl="??", string=str(self))


class ParaInputChannel(InputChannel):
    def __init__(self, ch_name, paras, var_name=None, is_str=False):
        super(ParaInputChannel, self).__init__(ch_name, var_name)
        assert all(isinstance(para, (int, float, str)) for para in paras)
        self.paras = paras
        self.is_str = is_str

    def __str__(self):
        result = str(self.ch_name)
        for para in self.paras:
            if isinstance(para, str) and self.is_str:
                result += "[\"" + str(para) + "\"]"
            else:
                result += "[" + str(para) + "]"
        result += "?"
        if self.var_name:
            result += str(self.var_name)
        result += ";"
        return result


class OutputChannel(HCSP):
    def __init__(self, ch_name, expr=None, meta=None):
        super(OutputChannel, self).__init__()
        self.type = "output_channel"
        if isinstance(ch_name, str):
            ch_name = Channel(ch_name)
        assert isinstance(ch_name, Channel)
        assert expr is None or isinstance(expr, Expr)
        self.ch_name = ch_name  # Channel
        self.expr = expr  # Expr or None
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.expr == other.expr and self.ch_name == other.ch_name

    def __repr__(self):
        if isinstance(self.expr, Expr):
            return "OutputC(%s,%s)" % (self.ch_name, self.expr)
        else:
            return "OutputC(%s)" % self.ch_name

    def __str__(self):
        if self.expr:
            return str(self.ch_name) + "!" + str(self.expr) + ";"
        else:
            return str(self.ch_name) + "!" + ";"

    def __hash__(self):
        return hash(("OutputChannel", self.ch_name, self.expr))

    def get_vars(self):
        if self.expr:
            return self.expr.get_vars()
        return set()

    def get_fun_names(self):
        if self.expr:
            return self.expr.get_fun_names()
        return set()

    def get_zero_arity_funcs(self):
        if self.expr:
            return self.expr.get_zero_arity_funcs()
        return set()

    def get_chs(self):
        return {self.ch_name}

    def sc_str(self):
        return re.sub(pattern="!", repl="!!", string=str(self))


class ParaOutputChannel(OutputChannel):
    def __init__(self, ch_name, paras, expr=None, is_str=False):
        super(ParaOutputChannel, self).__init__(ch_name, expr)
        assert all(isinstance(para, (int, float, str)) for para in paras)
        self.paras = paras
        self.is_str = is_str

    def __str__(self):
        result = str(self.ch_name)
        for para in self.paras:
            if isinstance(para, str) and self.is_str:
                result += "[\"" + str(para) + "\"]"
            else:
                result += "[" + str(para) + "]"
        if self.expr:
            result += str(self.expr)
        result += ";"
        return result


def is_comm_channel(hp):
    return hp.type == "input_channel" or hp.type == "output_channel"


class Sequence(HCSP):
    def __init__(self, *hps, meta=None):
        """hps is a list of hybrid programs."""
        super(Sequence, self).__init__()
        self.type = "sequence"
        assert all(isinstance(hp, HCSP) for hp in hps)
        assert len(hps) >= 1
        self.hps = []
        for hp in hps:
            if isinstance(hp, Sequence):
                self.hps.extend(hp.hps)
            else:
                self.hps.append(hp)
        self.hps = tuple(self.hps)
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.hps == other.hps

    def __repr__(self):
        return "Seq(%s)" % ", ".join(repr(hp) for hp in self.hps)

    def __str__(self):
        return " ".join(
            str(hp) if hp.priority() > self.priority() else "{" + str(hp) + "}"
            for hp in self.hps)

    def __hash__(self):
        return hash(("Sequence", self.hps))

    def sc_str(self):
        return "; ".join(
            hp.sc_str() if hp.priority() > self.priority() else "(" + hp.sc_str() + ")"
            for hp in self.hps)

    def get_vars(self):  # in assignments
        var_set = set()
        for hp in self.hps:
            # if isinstance(hp, Assign):
            # var_set = var_set.union(hp.get_vars())
            var_set.update(hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for hp in self.hps:
            fun_set.update(hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for hp in self.hps:
            fun_set.update(hp.get_zero_arity_funcs())
        return fun_set

    def get_chs(self):
        ch_set = set()
        for hp in self.hps:
            # ch_set = ch_set.union(hp.get_chs())
            ch_set.update(hp.get_chs())
        return ch_set

def seq(hps):
    """Returns the sequential composition of hps. Special case when hps has
    length 0 or 1.

    """
    assert all(isinstance(hp, HCSP) for hp in hps)
    hps = [hp for hp in hps if hp != Skip()]
    if len(hps) == 0:
        return Skip()
    elif len(hps) == 1:
        return hps[0]
    else:
        return Sequence(*hps)


class IChoice(HCSP):
    """Represents internal choice of the form P ++ Q ++ N."""
    def __init__(self, *hps, meta=None):
        super(IChoice, self).__init__()
        self.type = "ichoice"
        assert all(isinstance(hp, HCSP) for hp in hps)
        assert len(hps) >= 1
        self.hps = tuple(hps)
        # assert isinstance(hp1, HCSP) and isinstance(hp2, HCSP)
        # self.hp1 = hp1
        # self.hp2 = hp2
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.hps == other.hps

    def __str__(self):
        return " ++ ".join(str(hp) if hp.priority() > self.priority() else "(" + str(hp) + ")"
                for hp in self.hps)

    def __repr__(self):
        return "IChoice(%s)" % ", ".join(repr(hp) for hp in self.hps)

    def __hash__(self):
        return hash(("ICHOICE", self.hps))

    def get_vars(self):
        var_set = set()
        for hp in self.hps:
            var_set.update(hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for hp in self.hps:
            fun_set.update(hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for hp in self.hps:
            fun_set.update(hp.get_zero_arity_funcs())
        return fun_set

class ODE(HCSP):
    """Represents an ODE program of the form

    {F(s',s) = 0 & B} |> Q

    """
    def __init__(self, eqs, constraint, *, out_hp=Skip(), meta=None, inv=tuple()):
        """Each equation is of the form (var_name, expr), where var_name
        is the name of the variable, and expr is its derivative.

        constraint is a boolean formula.

        """
        super(ODE, self).__init__()
        assert isinstance(eqs, list)
        for eq in eqs:
            assert isinstance(eq, tuple) and len(eq) == 2
            assert isinstance(eq[0], str) and isinstance(eq[1], Expr)
        assert isinstance(constraint, Expr)
        assert isinstance(inv, tuple)
        for sub_inv in inv:
            assert isinstance(sub_inv, assertion.Assertion)
        assert not out_hp or isinstance(out_hp, HCSP)

        self.type = "ode"
        self.eqs = eqs  # list of tuples (string, Expr)
        self.constraint = constraint  # Expr
        self.out_hp = out_hp  # None or hcsp
        self.meta = meta
        self.inv = inv

    def __eq__(self, other):
        return self.type == other.type and self.eqs == other.eqs and \
               self.constraint == other.constraint and self.out_hp == other.out_hp

    def __str__(self):
        str_eqs = ", ".join(var_name + "_dot = " + str(expr) for var_name, expr in self.eqs)
        str_out_hp = "" if isinstance(self.out_hp, Skip) else " |> " + str(self.out_hp)
        return "{" + str_eqs + " & " + str(self.constraint) + "}" + str_out_hp

    def __repr__(self):
        if self.out_hp == Skip():
            return "ODE(%s, %s)" % (repr(self.eqs), repr(self.constraint))
        else:
            return "ODE(%s, %s, out_hp=%s)" % (repr(self.eqs), repr(self.constraint), repr(self.out_hp))

    def __hash__(self):
        return hash(("ODE", self.eqs, self.constraint, self.out_hp))

    def get_vars(self):
        var_set = set()
        for variable, expression in self.eqs:
            var_set.add(variable)
            var_set.update(expression.get_vars())
        var_set.update(self.constraint.get_vars())
        var_set.update(self.out_hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for _, expression in self.eqs:
            fun_set.update(expression.get_fun_names())
        fun_set.update(self.constraint.get_fun_names())
        fun_set.update(self.out_hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for _, expression in self.eqs:
            fun_set.update(expression.get_zero_arity_funcs())
        fun_set.update(self.constraint.get_zero_arity_funcs())
        fun_set.update(self.out_hp.get_zero_arity_funcs())
        return fun_set

    def get_chs(self):
        return self.out_hp.get_chs()


class ODE_Comm(HCSP):
    """Represents an ODE program that can be interrupted by
    communications, of the form

    {F(s',s) = 0 & B} |> [] (io_i --> Q_i)

    """
    def __init__(self, eqs, constraint, io_comms, meta=None):
        """Each equation is of the form (var_name, expr). Each element
        of io_comms is of the form (comm_hp, out_hp), where comm_hp
        is a communication process (either InputChannel or OutputChannel).
        Whenever one of comm_hp is received, the ODE stops and performs
        out_hp.

        """
        super(ODE_Comm, self).__init__()
        assert isinstance(eqs, list)
        for eq in eqs:
            assert isinstance(eq, tuple) and len(eq) == 2
            assert isinstance(eq[0], str) and isinstance(eq[1], Expr)
        assert isinstance(constraint, Expr)
        assert isinstance(io_comms, list)
        for io_comm in io_comms:
            assert isinstance(io_comm, tuple) and len(io_comm) == 2
            assert is_comm_channel(io_comm[0]) and isinstance(io_comm[1], HCSP)

        self.type = "ode_comm"
        self.eqs = eqs  # list
        self.constraint = constraint  # Expr
        self.io_comms = io_comms  # list
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.eqs == other.eqs and \
               self.constraint == other.constraint and self.io_comms == other.io_comms

    def __str__(self):
        str_eqs = ", ".join(var_name + "_dot = " + str(expr) for var_name, expr in self.eqs)
        str_io_comms = ", ".join(str(comm_hp)[:-1] + " --> " + str(out_hp) for comm_hp, out_hp in self.io_comms)
        return "{" + str_eqs + " & " + str(self.constraint) + "} |> [] (" + str_io_comms + ")"

    def __repr__(self):
        str_eqs = ", ".join(var_name + ", " + str(expr) for var_name, expr in self.eqs)
        str_io_comms = ", ".join(str(comm_hp)[:-1] + ", " + str(out_hp) for comm_hp, out_hp in self.io_comms)
        return "ODEComm(%s, %s, %s)" % (str_eqs, str(self.constraint), str_io_comms)

    def __hash__(self):
        return hash(("ODE_Comm", self.eqs, self.constraint, self.io_comms))

    def sc_str(self):
        derivatives = "DOT(" + ", ".join(var_name for var_name, _ in self.eqs) + ")"
        expressions = "{" + ", ".join(str(expr) for _, expr in self.eqs) + "}"
        domain = "DOMAIN(" + ("TRUE" if self.constraint == true_expr else str(self.constraint)) + ")"
        assert all(isinstance(out_hp, Var) for _, out_hp in self.io_comms)
        interupt = "INTERRUPT({" + ", ".join(comm_hp.sc_str() for comm_hp, _ in self.io_comms) + "}{" \
                   + ", ".join(out_hp.sc_str() for _, out_hp in self.io_comms) + "})"
        return derivatives + "=" + expressions + " " + domain + " " + interupt

    def get_vars(self):
        var_set = set()
        for variable, expression in self.eqs:
            var_set.add(variable)
            var_set.update(expression.get_vars())
        var_set.update(self.constraint.get_vars())
        for ch, out_hp in self.io_comms:
            var_set.update(ch.get_vars())
            var_set.update(out_hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for _, expression in self.eqs:
            fun_set.update(expression.get_fun_names())
        fun_set.update(self.constraint.get_fun_names())
        for ch, out_hp in self.io_comms:
            fun_set.update(ch.get_fun_names())
            fun_set.update(out_hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for _, expression in self.eqs:
            fun_set.update(expression.get_zero_arity_funcs())
        fun_set.update(self.constraint.get_zero_arity_funcs())
        for ch, out_hp in self.io_comms:
            fun_set.update(ch.get_zero_arity_funcs())
            fun_set.update(out_hp.get_zero_arity_funcs())
        return fun_set

    def get_chs(self):
        ch_set = set()
        for ch, out_hp in self.io_comms:
            ch_set.update(ch.get_chs())
            ch_set.update(out_hp.get_chs())
        return ch_set


class Loop(HCSP):
    """Represents an infinite loop of a program.
    
    hp : HCSP - body of the loop.
    constraint : Expr - loop condition, default to true.
    inv : tuple of OrdinaryAssertions - invariants

    """
    def __init__(self, hp, *, inv=tuple(), constraint=true_expr, meta=None):
        super(Loop, self).__init__()
        self.type = 'loop'
        assert isinstance(hp, HCSP)
        assert isinstance(inv, tuple)
        for sub_inv in inv:
            assert isinstance(sub_inv, assertion.OrdinaryAssertion)
        self.hp = hp
        self.inv = inv
        self.constraint = constraint
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.hp == other.hp and \
            self.constraint == other.constraint and self.inv == other.inv

    def __repr__(self):
        if self.constraint == true_expr:
            return "Loop(%s)" % (repr(self.hp))
        else:
            return "Loop(%s, %s)" % (repr(self.hp), str(self.constraint))

    def __str__(self):
        if self.constraint == true_expr:
            return "{%s}*" % str(self.hp)
        else:
            return "{%s}*(%s)" % (str(self.hp), str(self.constraint))

    def __hash__(self):
        return hash(("Loop", self.hp, self.constraint))

    def sc_str(self):
        assert isinstance(self.hp, Var)
        return self.hp.sc_str() + "**"

    def get_vars(self):
        return self.hp.get_vars()

    def get_fun_names(self):
        return self.hp.get_fun_names()

    def get_zero_arity_funcs(self):
        return self.hp.get_zero_arity_funcs()

    def get_chs(self):
        return self.hp.get_chs()

class Parallel(HCSP):
    def __init__(self, *hps, meta=None):
        """hps is a list of hybrid programs."""
        super(Parallel, self).__init__()
        self.type = "parallel"
        assert all(isinstance(hp, HCSP) for hp in hps)
        assert len(hps) >= 2
        # self.hps = list(hps)  # type(hps) == tuple
        self.hps = []
        for hp in hps:
            if hp.type == "parallel":
                self.hps.extend(hp.hps)
            else:
                self.hps.append(hp)
        self.hps = tuple(self.hps)
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.hps == other.hps

    def __repr__(self):
        return "Parallel(%s)" % (", ".join(repr(hp) for hp in self.hps))

    def __str__(self):
        return " || ".join(
            str(hp) if hp.priority() > self.priority() else "{" + str(hp) + "}"
            for hp in self.hps)

    def __hash__(self):
        return hash(("Parallel", self.hps))

    def sc_str(self):
        return " || ".join(
            hp.sc_str() if hp.priority() > self.priority() else "(" + hp.sc_str() + ")"
            for hp in self.hps)

    def get_vars(self):
        var_set = set()
        for hp in self.hps:
            var_set.update(hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for hp in self.hps:
            fun_set.update(hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for hp in self.hps:
            fun_set.update(hp.get_zero_arity_funcs())
        return fun_set


    def get_chs(self):
        ch_set = set()
        for hp in self.hps:
            ch_set.update(hp.get_chs())
        return ch_set


class SelectComm(HCSP):
    def __init__(self, *io_comms, meta=None):
        """io_comms is a list of pairs, where the first element of each
        pair must be a communication, and the second entry can be
        any program.
        
        """
        super(SelectComm, self).__init__()
        self.type = "select_comm"
        assert len(io_comms) >= 2

        assert all(is_comm_channel(comm_hp) and isinstance(out_hp, HCSP)
                   for comm_hp, out_hp in io_comms)
        self.io_comms = tuple(io_comms)
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.io_comms == other.io_comms

    def __repr__(self):
        return "SelectComm(%s)" % (",".join("%s,%s" % (repr(comm_hp), repr(out_hp))
                                            for comm_hp, out_hp in self.io_comms))

    def __str__(self):
        return " $ ".join(
            "%s --> %s" % (str(comm_hp)[:-1], out_hp) if out_hp.priority() > self.priority() else
            "%s --> {%s}" % (str(comm_hp)[:-1], out_hp)
            for comm_hp, out_hp in self.io_comms)

    def __hash__(self):
        return hash(("SelectComm", self.io_comms))

    def sc_str(self):
        return " [[ ".join("(%s; %s)" % (comm_hp.sc_str(), out_hp.sc_str()) if out_hp.priority() > self.priority()
                           else "(%s; (%s))" % (comm_hp.sc_str(), out_hp.sc_str()) for comm_hp, out_hp in self.io_comms)

    def get_vars(self):
        var_set = set()
        for ch, out_hp in self.io_comms:
            var_set.update(ch.get_vars())
            var_set.update(out_hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for ch, out_hp in self.io_comms:
            fun_set.update(ch.get_fun_names())
            fun_set.update(out_hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for ch, out_hp in self.io_comms:
            fun_set.update(ch.get_zero_arity_funcs())
            fun_set.update(out_hp.get_zero_arity_funcs())
        return fun_set

    def get_chs(self):
        ch_set = set()
        for ch, out_hp in self.io_comms:
            ch_set.update(ch.get_chs())
            ch_set.update(out_hp.get_chs())
        return ch_set


class ITE(HCSP):
    def __init__(self, if_hps, else_hp=None, meta=None):
        """if-then-else statements.

        if_hps : List[Tuple[Expr, HCSP]] - list of condition-program pairs.
        else_hp : [None, HCSP]

        The program associated to the first true condition in if_hps will
        be executed. If no condition is true, else_hp is executed.

        """
        super(ITE, self).__init__()
        assert all(isinstance(cond, Expr) and isinstance(hp, HCSP) for cond, hp in if_hps)
        assert len(if_hps) > 0, "ITE: must have at least one if branch"
        assert else_hp is None or isinstance(else_hp, HCSP)
        self.type = "ite"
        self.if_hps = tuple(tuple(p) for p in if_hps)
        self.else_hp = else_hp
        self.meta = meta

    def __eq__(self, other):
        return self.type == other.type and self.if_hps == other.if_hps and self.else_hp == other.else_hp

    def __repr__(self):
        if_hps_strs = ", ".join("%s, %s" % (cond, repr(hp)) for cond, hp in self.if_hps)
        return "ITE(%s, %s)" % (if_hps_strs, repr(self.else_hp))

    def __str__(self):
        res = "if (%s) { %s " % (self.if_hps[0][0], self.if_hps[0][1])
        for cond, hp in self.if_hps[1:]:
            res += "} else if (%s) { %s " % (cond, hp)
        if self.else_hp is None:
            res += "}"
        else:
            res += "} else { %s }" % self.else_hp
        return res

    def __hash__(self):
        return hash(("ITE", self.if_hps, self.else_hp))

    def sc_str(self):
        assert len(self.if_hps) == 1 and isinstance(self.if_hps[0][1], Var) and (self.else_hp is None or isinstance(self.else_hp, Var))
        if self.else_hp is None:
            return "if %s then %s" % (self.if_hps[0][0], self.if_hps[0][1].sc_str())
        else:
            return "if %s then %s else %s" % (self.if_hps[0][0], self.if_hps[0][1].sc_str(), self.else_hp.sc_str())

    def get_vars(self):
        var_set = set()
        for cond, hp in self.if_hps:
            var_set.update(cond.get_vars())
            var_set.update(hp.get_vars())
        if self.else_hp is not None:
            var_set.update(self.else_hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for cond, hp in self.if_hps:
            fun_set.update(cond.get_fun_names())
            fun_set.update(hp.get_fun_names())
        if self.else_hp is not None:
            fun_set.update(self.else_hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for cond, hp in self.if_hps:
            fun_set.update(cond.get_zero_arity_funcs())
            fun_set.update(hp.get_zero_arity_funcs())
        if self.else_hp is not None:
            fun_set.update(self.else_hp.get_zero_arity_funcs())
        return fun_set

    def get_chs(self):
        ch_set = set()
        for _, hp in self.if_hps:
            ch_set.update(hp.get_chs())
        if self.else_hp is not None:
            ch_set.update(self.else_hp.get_chs())
        return ch_set

def ite(if_hps, else_hp):
    """Construction of if-then-else with simplifications."""
    new_if_hps, new_else_hp = [], else_hp
    for cond, if_hp in if_hps:
        if cond == true_expr:
            new_else_hp = if_hp
            break
        elif cond == false_expr:
            continue
        else:
            new_if_hps.append((cond, if_hp))
    if len(new_if_hps) == 0:
        if new_else_hp is None:
            return Skip()
        return new_else_hp
    else:
        return ITE(new_if_hps, new_else_hp)

def get_comm_chs(hp):
    """Returns the list of communication channels for the given program.
    
    Result is a list of pairs (ch_name, '?'/'!').
    
    """
    assert isinstance(hp, HCSP)
    collect = []

    def rec(_hp):
        if _hp.type == 'input_channel':
            collect.append((_hp.ch_name, '?'))
        elif _hp.type == 'output_channel':
            collect.append((_hp.ch_name, '!'))
        elif _hp.type == 'sequence':
            for arg in _hp.hps:
                rec(arg)
        elif _hp.type == 'ode':
            if _hp.out_hp:
                rec(_hp.out_hp)
        elif _hp.type in ('ode_comm', 'select_comm'):
            for comm_hp, out_hp in _hp.io_comms:
                rec(comm_hp)
                rec(out_hp)
        elif _hp.type == 'loop':
            rec(_hp.hp)
        elif _hp.type == 'ite':
            for _, sub_hp in _hp.if_hps:
                rec(sub_hp)
            if _hp.else_hp is not None:
                rec(_hp.else_hp)
    
    rec(hp)
    return list(OrderedDict.fromkeys(collect))

class HoareTriple:
    """A program with pre- and post-conditions"""
    def __init__(self, pre, hp, post, functions=None, meta=None):
        self.pre = list(pre)
        self.post = list(post)
        assert all(isinstance(subpost, assertion.OrdinaryAssertion) for subpost in post)
        self.hp = hp
        if functions is None:
            functions = dict()
        self.functions = functions
        self.meta = meta

class Procedure:
    """Declared procedure within a process."""
    def __init__(self, name, hp, meta=None):
        self.name = name
        if isinstance(hp, str):
            from ss2hcsp.hcsp.parser import hp_parser
            hp = hp_parser.parse(hp)
        self.hp = hp
        self.meta = meta

    def __eq__(self, other):
        return self.name == other.name and self.hp == other.hp

    def __repr__(self):
        return "Procedure(%s,%s)" % (self.name, repr(self.hp))

    def __str__(self):
        return "procedure %s begin %s end" % (self.name, str(self.hp))


class HCSPInfo:
    """HCSP process with extra information."""
    def __init__(self, name, hp, *, outputs=None, procedures=None):
        self.name = name
        self.hp = hp
        
        # List of output variables, None indicates output everything
        self.outputs = outputs

        # List of declared procedures
        if procedures is None:
            procedures = []
        self.procedures = procedures

    def __str__(self):
        res = self.name + ' ::=\n'
        for procedure in self.procedures:
            res += str(procedure) + '\n'
        res += str(self.hp)
        return res

    def __repr__(self):
        return "HCSPInfo(%s, %s)" % (self.name, str(self.hp))

    def __eq__(self, other):
        return self.name == other.name and self.hp == other.hp

def subst_comm_all(hp, inst):
    """Perform all substitutions given in inst. Detect cycles.
    
    First compute a topological sort of dependency in inst, which will
    provide the order of substitution.

    """
    # Set of all variables to be substituted
    all_vars = set(inst.keys())

    # Mapping variable to its dependencies
    dep_order = dict()
    for var in all_vars:
        dep_order[var] = list(inst[var].get_vars().intersection(all_vars))

    topo_order = topological_sort(dep_order)
    for var in reversed(topo_order):
        hp = hp.subst_comm({var: inst[var]})
    return hp

def HCSP_subst(hp, inst):
    """Substitution of program variables for their instantiations."""
    assert isinstance(hp, HCSP)
    if isinstance(hp, Var):
        if hp.name in inst:
            return inst[hp.name]
        else:
            return hp
    elif isinstance(hp, (Skip, Wait, Assign, Assert, Test, Log, InputChannel, OutputChannel)):
        return hp
    elif isinstance(hp, Loop):
        hp.hp = HCSP_subst(hp.hp, inst)
        return hp
    elif isinstance(hp, Sequence):
        hps = [HCSP_subst(sub_hp, inst) for sub_hp in hp.hps]
        return Sequence(*hps)
    elif isinstance(hp, ODE):
        hp.out_hp = HCSP_subst(hp.out_hp, inst)
        return hp
    elif isinstance(hp, (ODE_Comm, SelectComm)):
        hp.io_comms = tuple((io_comm[0], HCSP_subst(io_comm[1], inst)) for io_comm in hp.io_comms)
        return hp
    elif isinstance(hp, ITE):
        hp.if_hps = tuple((e[0], HCSP_subst(e[1], inst)) for e in hp.if_hps)
        if hp.else_hp is not None:
            hp.else_hp = HCSP_subst(hp.else_hp, inst)
        return hp
    else:
        print(hp)
        raise NotImplementedError

def reduce_procedures(hp, procs, strict_protect=None):
    """Reduce number of procedures in the process by inlining.

    hp : HCSP - input process.
    procs : Dict[str, HCSP] - mapping from procedure name to procedure body.
    strict_protect : Set[str] - this set of procedures will never be inlined.

    The algorithm for finding which procedures to inline is partly heuristic,
    with settings specified by the various flags. It performs the following steps:

    1. Traverse the procedure definitions to find dependency relations.

    2. Inline any procedure that does not depend on any other procedure (except
       those that are strictly protected).

    3. When step 2 can not find any procedure to inline, remove one of the
       procedures from consideration.

    This function directly modifies procs, and returns the new HCSP process.

    """
    if strict_protect is None:
        strict_protect = set()

    # First, construct the dependency relation. We use the empty string
    # to represent the toplevel process.
    dep_relation = dict()
    dep_relation[""] = hp.get_contain_hps()
    for name, proc in procs.items():
        dep_relation[name] = proc.get_contain_hps()

    # Construct reverse dependency relation, for heuristic selection of
    # inlined functions. For each hp
    rev_dep_count = dict()
    for name in procs:
        rev_dep_count[name] = 0
    for name, count in hp.get_contain_hps_count().items():
        rev_dep_count[name] += count
    for _, proc in procs.items():
        for name, count in proc.get_contain_hps_count().items():
            rev_dep_count[name] += count

    # Set of procedures to be inlined, in order of removal
    inline_order = []

    # Set of procedures that are kept, initially the toplevel process
    # and the strictly protected processes.
    fixed = {""}.union(strict_protect)

    # Also protect procedures that are too large and invoked too many times
    for name, proc in procs.items():
        if rev_dep_count[name] >= 2 and proc.size() > 8:
            fixed.add(name)

    # Get the order of inlining
    while True:
        # First, find procedures that only depend on inlined or fixed procedures
        found = False
        for name, dep in dep_relation.items():
            if name not in fixed and name not in inline_order and \
                dep.issubset(set(inline_order).union(fixed)):
                found = True
                inline_order.append(name)

        # If all procedures are either inlined or fixed, exit the loop
        if len(fixed) + len(inline_order) == len(dep_relation):
            break

        # If some procedure is found, repeat the iteration
        if found:
            continue

        # No procedure is found, move one procedure to fixed
        for name in dep_relation:
            if name not in fixed and name not in inline_order:
                fixed.add(name)
                break
        
    # Next, recursively perform substitutions
    for inline_name in inline_order:
        for name, dep in dep_relation.items():
            if name == "" and inline_name in dep:
                hp = HCSP_subst(hp, {inline_name: procs[inline_name]})
            elif inline_name in dep:
                procs[name] = HCSP_subst(procs[name], {inline_name: procs[inline_name]})
        del procs[inline_name]

    return hp


class HCSPProcess:
    """System of HCSP processes. Input is a list of (name, HCSP) pairs."""
    def __init__(self, hps=None):
        """Initialize with an optional list of definitions."""

        # List of (name, HCSP) pairs.
        self.hps = []
        if hps:
            for name, hp in hps:
                self.hps.append((name, hp))

    def add(self, name, hp):
        """Insert (name, hp) at the end."""
        self.hps.append((name, hp))

    def extend(self, lst):
        """Insert list of (name, hp) at the end."""
        if isinstance(lst, HCSPProcess):
            self.hps.extend(lst.hps)
        else:
            self.hps.extend(lst)

    def insert(self, n, name, hp):
        """Insert (name, hp) at position n."""
        self.hps.insert(n, (name, hp))

    def substitute(self):
        """Substitute program variables for their definitions."""
        def _substitute(_hp):
            assert isinstance(_hp, HCSP)
            if isinstance(_hp, Var):
                _name = _hp.name
                if _name in substituted.keys():
                    _hp = substituted[_name]
                elif _name in hps_dict:
                    _hp = _substitute(hps_dict[_name])
                    substituted[_name] = _hp
                    del hps_dict[_name]
            elif isinstance(_hp, Loop):
                _hp.hp = _substitute(_hp.hp)
            elif isinstance(_hp, Sequence):
                _hps = [_substitute(sub_hp) for sub_hp in _hp.hps]
                _hp = Sequence(*_hps)
                # _hp = Sequence(tuple(_substitute(sub_hp) for sub_hp in _hp.hps))
            elif isinstance(_hp, ODE):
                _hp.out_hp = _substitute(_hp.out_hp)
            elif isinstance(_hp, (ODE_Comm, SelectComm)):
                _hp.io_comms = tuple((io_comm[0], _substitute(io_comm[1])) for io_comm in _hp.io_comms)
            elif isinstance(_hp, ITE):
                _hp.if_hps = tuple((e[0], _substitute(e[1])) for e in _hp.if_hps)
                if _hp.else_hp is not None:
                    _hp.else_hp = _substitute(_hp.else_hp)
            return _hp

        hps_dict = {name: hp for name, hp in self.hps}
        #assert len(hps_dict) == len(self.hps)
        substituted = dict()
        while hps_dict:
            name, hp = hps_dict.popitem()
            assert name not in substituted
            substituted[name] = _substitute(hp)
        assert set(substituted.keys()) == set(name for name, _ in self.hps)

        return substituted

    def __str__(self):
        return "\n".join("%s ::= %s" % (name, str(hp)) for name, hp in self.hps)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.hps == other.hps

    def sc_str(self):
        assert len([_ for _, hp in self.hps if isinstance(hp, Parallel)]) == 1
        system, process = None, None
        for name, hp in self.hps:
            if isinstance(hp, Parallel):
                system, process = name, hp
                break
        sys_def = "systemDef " + system + " ::= " + process.sc_str() + "\n\n"
        var_def = "variableDef ::= " + ("; ".join(self.get_vars())) + "\n\n"
        ch_def = "channelDef ::= " + ("; ".join([ch.name for ch in self.get_chs()])) + "\n\n"
        hp_def = "\n".join("processDef " + "%s ::= %s" % (name, hp.sc_str()) for name, hp in self.hps
                           if not isinstance(hp, Parallel))
        return sys_def + var_def + ch_def + hp_def

    def get_vars(self):
        var_set = set()
        for _, hp in self.hps:
            var_set.update(hp.get_vars())
        return var_set

    def get_fun_names(self):
        fun_set = set()
        for _, hp in self.hps:
            fun_set.update(hp.get_fun_names())
        return fun_set

    def get_zero_arity_funcs(self):
        fun_set = set()
        for _, hp in self.hps:
            fun_set.update(hp.get_zero_arity_funcs())
        return fun_set

    def get_chs(self):
        ch_set = set()
        for _, hp in self.hps:
            ch_set.update(hp.get_chs())
        return ch_set



"""Unit test for parsing files."""

import unittest

from ss2hcsp.hcsp.hcsp import HCSPInfo, Procedure
from ss2hcsp.hcsp import module
from ss2hcsp.hcsp.parser import hp_parser, module_parser, parse_file, parse_module_file, \
                                expr_parser


class ParserTest(unittest.TestCase):
    def testParseFile(self):
        infos = parse_file("""
            P0 ::= x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*
            P1 ::= { wait(2); p2c?x; c2p!x-1; }*
        """)

        self.assertEqual(infos, [
            HCSPInfo("P0", hp_parser.parse("x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*")),
            HCSPInfo("P1", hp_parser.parse("{ wait(2); p2c?x; c2p!x-1; }*"))
        ])

    def testParseModuleFile(self):
        infos = parse_module_file("""
            %type: module

            module P0():
            output x;
            begin
                x := 0;
                { {x_dot = 1 & true} |> [](p2c!x --> skip; ) c2p?x; }*
            end
            endmodule

            module P1():
            begin
                { wait(2); p2c?x; c2p!x-1; }*
            end
            endmodule

            system
            P0() ||
            P1()
            endsystem
        """)

        self.assertEqual(infos, [
            HCSPInfo("P0", hp_parser.parse("x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*")),
            HCSPInfo("P1", hp_parser.parse("{ wait(2); p2c?x; c2p!x-1; }*"))
        ])

    def testParseModuleFile2(self):
        infos = parse_module_file("""
            %type: module

            module P0(p2c,c2p):
            output x;
            begin
              x := 0;
              {{x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*
            end
            endmodule

            module P1(p2c,c2p):
            begin
              { wait(2); p2c?x; c2p!x-1; }*
            end
            endmodule

            system
              P0a = P0(ch1,ch2) ||
              P1a = P1(ch1,ch2) ||
              P0b = P0(ch3,ch4) ||
              P1b = P1(ch3,ch4)
            endsystem
        """)

        self.assertEqual(infos, [
            HCSPInfo("P0a", hp_parser.parse("x := 0; { {x_dot = 1 & true} |> [](ch1!x --> skip;) ch2?x; }*")),
            HCSPInfo("P1a", hp_parser.parse("{ wait(2); ch1?x; ch2!x-1; }*")),
            HCSPInfo("P0b", hp_parser.parse("x := 0; { {x_dot = 1 & true} |> [](ch3!x --> skip;) ch4?x; }*")),
            HCSPInfo("P1b", hp_parser.parse("{ wait(2); ch3?x; ch4!x-1; }*")),
        ])

    def testParseModule3(self):
        infos = parse_module_file("""
            %type: module

            module P0(i):
            output x;
            begin
              x := 0;
              { {x_dot = 1 & true} |> [](p2c[i]!x --> skip; ) c2p[i]?x; }*
            end
            endmodule

            module P1(i):
            begin
              { wait(2); p2c[i]?x; c2p[i]!x-1; }*
            end
            endmodule

            system
              P0a=P0($0) || P1a=P1($0) ||
              P0b=P0($1) || P1b=P1($1)
            endsystem
        """)

        self.assertEqual(infos, [
            HCSPInfo("P0a", hp_parser.parse("x := 0; { {x_dot = 1 & true} |> [](p2c[0]!x --> skip;) c2p[0]?x; }*")),
            HCSPInfo("P1a", hp_parser.parse("{ wait(2); p2c[0]?x; c2p[0]!x-1; }*")),
            HCSPInfo("P0b", hp_parser.parse("x := 0; { {x_dot = 1 & true} |> [](p2c[1]!x --> skip;) c2p[1]?x; }*")),
            HCSPInfo("P1b", hp_parser.parse("{ wait(2); p2c[1]?x; c2p[1]!x-1; }*")),
        ])

    def testParseProcedure(self):
        mod = module_parser.parse("""
            module P0():
            procedure incr begin
              x := x+1;
            end
            begin
              x := 0; @incr; @incr; ch!x;
            end
            endmodule
        """)

        self.assertEqual(mod, module.HCSPModule(
            "P0", "x := 0; @incr; @incr; ch!x;",
            procedures=[Procedure("incr", "x := x+1;")],
        ))

    def testQuantifierExpr(self):
        exprs = [expr_parser.parse("\\forall x. x != 0 -> 1/x != 0"),
                 expr_parser.parse("\\forall {x, y}. x^2 + y^2 >= 0"),
                 expr_parser.parse("\\exists x. x > 0 -> x * y^2 == 1"),
                 expr_parser.parse("\\exists {x, y}. x^2 + y^2 == 0"),
                 expr_parser.parse("\\forall x. \\exists y. x > 0 -> x * y^2 == 1"),
                 expr_parser.parse("\\exists y. \\forall x. x > 0 -> x * y^2 == 1"),
                 expr_parser.parse("\\forall x. \\forall y. x^2 + y^2 >= 0")]

        for expr in exprs:
            print(expr) 

    def testNotExpr(self):
        exprs = [expr_parser.parse("!(x >= 0)")]

        for expr in exprs:
            print(expr)      


if __name__ == "__main__":
    unittest.main()



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import ArrayIdxExpr, RelExpr, AVar, AConst, true_expr, LogicExpr, conj, OpExpr, FunExpr
import ss2hcsp.hcsp.hcsp as hp


class Subsystem(SL_Block):
    """Subsystem is block in which there is a diagram"""
    def __init__(self, type, name, num_src, num_dest, st):
        super(Subsystem, self).__init__(type, name, num_src, num_dest, st)
        self.diagram = None

    def __str__(self):
        res = "%s: subsystem\n" % (self.name)
        res += "\n".join("  " + line for line in str(self.diagram).split('\n'))
        return res

    def __repr__(self):
        return "Subsystem(%s, %s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.num_src, self.num_dest, self.st, str(self.dest_lines), str(self.src_lines))

    def add_enabled_condition_to_innerBlocks(self, init_en_cond=true_expr):
        en_line = self.dest_lines[-1]
        en_cond = RelExpr(">", AVar(en_line.name), AConst(0))
        if init_en_cond != true_expr:
            en_cond = LogicExpr("&&", init_en_cond, en_cond)
        for block in self.diagram.blocks:
            if hasattr(block, "enable"):
                block.enable = en_cond
            elif block.type in ["enabled_subsystem", "triggered_subsystem"]:
                block.add_enabled_condition_to_innerBlocks(init_en_cond=en_cond)


class Triggered_Subsystem(Subsystem):
    def __init__(self, name, num_src, num_dest, trigger_type):
        super(Triggered_Subsystem, self).__init__("triggered_subsystem", name, num_src, num_dest, st=-1)

        # Trigger type
        assert trigger_type in ("rising", "falling", "either", None), \
            "Unknown trigger type: %s" % trigger_type
        self.trigger_type = trigger_type

        # A flag variable denoting if the subsystem was triggered at the last step
        self.triggered = self.name + "_triggered"

    def __str__(self):
        trig_var = self.dest_lines[-1].name
        res = "%s: on_%s(%s):\n" % (self.name, self.trigger_type, trig_var)
        res += "\n".join("  " + line for line in str(self.diagram).split('\n'))
        return res

    def __repr__(self):
        return "Triggered_Subsystem(%s, %s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.num_src, self.num_dest, self.trigger_type,
             str(self.dest_lines), str(self.src_lines))

    def get_pre_cur_trig_signals(self, index=None):
        """Obtain variables for current and previous trigger value."""
        line = self.dest_lines[-1].name
        cur_sig = AVar(line)
        pre_sig = AVar("pre_" + line)
        if index is not None:
            assert isinstance(index, int)
            cur_sig = ArrayIdxExpr(cur_sig, AConst(index))
            pre_sig = ArrayIdxExpr(pre_sig, AConst(index))
        return pre_sig, cur_sig

    def get_output_hp(self):
        output_hps = []
        if self.type == "stateflow":
            trigger_events = [(trigger_type, event) for trigger_type, event in self.input_events
                              if trigger_type in ['rising', 'falling', 'either']]
            if trigger_events:
                # Stateflow has trigger events, then the last dest_lines
                # is the trigger line
                for i, (trigger_type, event) in enumerate(trigger_events):
                    if len(trigger_events) == 1:
                        pre_sig, cur_sig = self.get_pre_cur_trig_signals()
                    else:
                        pre_sig, cur_sig = self.get_pre_cur_trig_signals(i)
                    trig_cond = self.get_discrete_triggered_condition(pre_sig, cur_sig, trigger_type)
                    push_event = hp.Assign(var_name=self.name+"EL",
                                           expr=FunExpr(fun_name="push", exprs=[AVar(self.name+"EL"), AConst(event)]))
                    execute_chart = hp.Var(self.exec_name)
                    pop_event = hp.Assign(var_name=self.name+"EL",
                                          expr=FunExpr(fun_name="pop", exprs=[AVar(self.name+"EL")]))
                    main_execute = hp.Sequence(push_event, execute_chart, pop_event)

                    pre_sig_all, cur_sig_all = self.get_pre_cur_trig_signals()
                    output_hps.append(hp.ITE([(trig_cond, main_execute)]))
                output_hps.append(hp.Assign(pre_sig_all.name, cur_sig_all))
                return hp.seq(output_hps)

            elif not self.input_events:
                assert self.st > 0
                return hp.Var(self.exec_name)

            else:
                return hp.Skip()

        elif self.type == "triggered_subsystem":
            pre_sig, cur_sig = self.get_pre_cur_trig_signals()
            trig_cond = self.get_discrete_triggered_condition(pre_sig, cur_sig, self.trigger_type)
            main_execute = hp.Var(self.name)
            return hp.seq([hp.ITE([(trig_cond, main_execute)]),
                            hp.Assign(pre_sig, cur_sig)])

        else:
            raise NotImplementedError

    def get_init_hps(self):
        # Initialize the triggered signals
        init_hps = list()
        if self.type == "stateflow":
            trigger_events = [(trigger_type, event) for trigger_type, event in self.input_events
                              if trigger_type in ['rising', 'falling', 'either']]
            if trigger_events:
                pre_sig_all, _ = self.get_pre_cur_trig_signals()
                if len(trigger_events) == 1:
                    init_hps.append(hp.Assign(pre_sig_all, AConst(0)))
                else:
                    init_hps.append(hp.Assign(pre_sig_all, AConst([0] * len(trigger_events))))
        elif self.type == "triggered_subsystem":
            pre_sig, _ = self.get_pre_cur_trig_signals()
            init_hps.append(hp.Assign(pre_sig, AConst(0)))
        else:
            raise NotImplementedError

        # Initialize the output variables
        for lines in self.src_lines:
            out_var = lines[0].name
            init_hps.append(hp.Assign(var_name=out_var, expr=AConst(0)))

        # Initialize the variables of the inner blocks
        if self.type == "triggered_subsystem":
            for block in self.diagram.blocks_dict.values():
                if block.type == "constant":
                    out_var = block.src_lines[0][0].name
                    init_hps.append(hp.Assign(var_name=out_var, expr=AConst(block.value)))
                elif block.type == "unit_delay":
                    init_hps.append(hp.Assign(var_name=block.name+"_state", expr=AConst(block.init_value)))
                elif block.type == "triggered_subsystem":
                    init_hps.extend(block.get_init_hps())
        return init_hps

    def get_discrete_triggered_condition(self, pre_sig, cur_sig, trigger_type):
        """Obtain the discrete trigger condition."""
        # Compare pre_sig and cur_sig with zero. Different comparisons for
        # different trigger conditions.
        if trigger_type == "rising":
            # (pre_sig < 0 && cur_sig >= 0) || (pre_sig == 0 && cur_sig > 0)
            op0, op1, op2, op3 = "<", ">=", "==", ">"
        elif trigger_type == "falling":
            # (pre_sig > 0 && cur_sig <= 0) || (pre_sig == 0 && cur_sig < 0)
            op0, op1, op2, op3 = ">", "<=", "==", "<"
        elif trigger_type == "either":
            # (pre_sig < 0 && cur_sig >= 0) || (pre_sig >= 0 && cur_sig < 0)
            op0, op1, op2, op3 = "<=", ">", ">", "<="
        else:
            raise NotImplementedError("Unknown trigger type: %s" % trigger_type)

        return LogicExpr(
            "||", LogicExpr("&&", RelExpr(op0, pre_sig, AConst(0)), RelExpr(op1, cur_sig, AConst(0))),
                 LogicExpr("&&", RelExpr(op2, pre_sig, AConst(0)), RelExpr(op3, cur_sig, AConst(0))))

    def get_continuous_triggered_condition(self):
        """Obtain the continuous trigger condition.
        
        Currently, we assume that the trigger line is the output of an
        integrator.

        """
        trig_line = self.dest_lines[-1]
        trig_sig = AVar(trig_line.name)
        assert trig_line.src_block.type == "integrator"
        trig_sig_dot = AVar(trig_line.src_block.dest_lines[0].name)

        if self.trigger_type == "rising":
            return conj(RelExpr("!=", AVar(self.triggered), AConst(1)),
                        RelExpr(">", trig_sig_dot, AConst(0)), RelExpr("==", trig_sig, AConst(0)))
        elif self.trigger_type == "falling":
            return conj(RelExpr("!=", AVar(self.triggered), AConst(1)),
                        RelExpr("<", trig_sig_dot, AConst(0)), RelExpr("==", trig_sig, AConst(0)))
        elif self.trigger_type == "either":
            return conj(RelExpr("!=", AVar(self.triggered), AConst(1)),
                        RelExpr("!=", trig_sig_dot, AConst(0)), RelExpr("==", trig_sig, AConst(0)))
        else:
            raise RuntimeError("Not implemented yet")

    def delete_subsystems(self):
        self.diagram.delete_subsystems()
        for block in self.diagram.block_dict.values():
            assert block.type not in ["subsystem", "enabled_subsystem"]
            if block.type == "triggered_subsystem":
                block.delete_subsystems()

    def get_procedures(self):
        """ Get the procedure body by topological sort. """
        # Get a list of sorted blocks
        block_dict = {name: block for name, block in self.diagram.blocks_dict.items()
                      if block.type not in ["in_port", "out_port", "constant"]}

        # Get a topological sort of the blocks
        # First, move all the Unit_Delay blocks from block_dict to sorted_blocks
        sorted_blocks = [block for block in block_dict.values() if block.type == "unit_delay"]
        for block in sorted_blocks:
            del block_dict[block.name]
        # Get a topological sort of the remaining blocks
        while block_dict:
            head_block_names = []
            for name, block in block_dict.items():
                src_blocks = block.get_src_blocks()
                if src_blocks.isdisjoint(set(block_dict.keys())):
                    sorted_blocks.append(block)
                    head_block_names.append(name)
            assert head_block_names
            for name in head_block_names:
                del block_dict[name]

        # Get the OUTPUT of each block in sorted_blocks
        output_hps = [block.get_output_hp() for block in sorted_blocks]
        # Get the UPDATE of Unit_Delay blocks
        update_hps = [block.get_update_hp() for block in sorted_blocks if block.type == "unit_delay"]

        result_hps = output_hps + update_hps
        result_hp = hp.Sequence(*result_hps) if len(result_hps) > 1 else result_hps[0]
        procedures = [hp.Procedure(name=self.name, hp=result_hp)]

        # Get the var-procedure mappings of the inner triggered subsystems
        for block in sorted_blocks:
            if block.type == "triggered_subsystem":
                inner_procedures = block.get_procedures()
                assert set(proc.name for proc in procedures).isdisjoint(
                    set(proc.name for proc in inner_procedures))
                procedures.extend(inner_procedures)

        return procedures


class Enabled_Subsystem(Subsystem):
    def __init__(self, name, num_src, num_dest):
        super(Enabled_Subsystem, self).__init__("enabled_subsystem", name, num_src, num_dest, st=-1)

    def __str__(self):
        en_line = self.dest_lines[-1]
        en_cond = RelExpr(">", AVar(en_line.name), AConst(0))
        res = "%s: on %s:\n" % (self.name, en_cond)
        res += "\n".join("  " + line for line in str(self.diagram).split('\n'))
        return res

    def __repr__(self):
        return "Enabled_Subsystem(%s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.num_src, self.num_dest, str(self.dest_lines), str(self.src_lines))



import os

def run_cmds():
    cmds = [
        'python -m unittest ss2hcsp.tests.expr_test',
        'python -m unittest ss2hcsp.tests.isabelle_test',
        'python -m unittest ss2hcsp.tests.matlab_test',
        'python -m unittest ss2hcsp.tests.module_test',
        'python -m unittest ss2hcsp.tests.optimize_test',
        'python -m unittest ss2hcsp.tests.parser_test',
        'python -m unittest ss2hcsp.tests.pprint_test',
        'python -m unittest ss2hcsp.tests.sf_convert_test',
        'python -m unittest ss2hcsp.tests.sf_isabelle_test',
        'python -m unittest ss2hcsp.tests.sim_test',
        'python -m unittest ss2hcsp.tests.simulator_test',
        'python -m unittest ss2hcsp.tests.systemC_test',
        'python -m unittest ss2hcsp.tests.topsort_test',
        'python -m unittest ss2hcsp.tests.transition_test',
    ]

    for cmd in cmds:
        os.system(cmd)

if __name__ == '__main__':
    run_cmds()


from ss2hcsp.hcsp import simulator
import json
from ss2hcsp.hcsp import parser
from ss2hcsp.hcsp import hcsp
from ss2hcsp.hcsp import pprint
import random
import time

path = "Examples\AADL\water_tank\system.txt"
file1 = open(path, "r", encoding='utf-8')
text = file1.read()
file1.close()

def raise_error(err_str):
    return json.dumps({
        'error': err_str
    })

try:
    if text.startswith('%type: module'):
        infos = parser.parse_module_file(text)
    else:
        infos = parser.parse_file(text)
except parser.ParseFileException as e:
   raise_error(e.error_msg)

sim_infos = []
hcsp_info = []
for info in infos:
    name, hp, procs = info.name, info.hp, info.procedures
    if hp.type == 'parallel':
        assert all(sub_hp.type == 'var' for sub_hp in hp.hps)
        hcsp_info.append({
            'name': name,
            'parallel': [sub_hp.name for sub_hp in hp.hps]
        })
    else:
        sim_infos.append(simulator.SimInfo(
            name, hp, outputs=info.outputs, procedures=info.procedures))
        lines, mapping = pprint.pprint_lines(hp, record_pos=True)
        json_procs = []
        for proc in info.procedures:
            #proc_lines, proc_mapping = pprint.pprint_lines(
            #    proc.hp, record_pos=True)
            json_procs.append({
                'name': proc.name,
                'text': str(proc.hp),
                # 'lines': proc_lines,
                # 'mapping': proc_mapping
            })
        hcsp_info.append({
            'name': name,
            'text': str(hp),
            'outputs': info.outputs,
            # 'lines': lines,
            # 'mapping': mapping,
            'procedures': json_procs
        })

num_steps = 200
num_show = 200
start_event = None
show_interval = None


def convert_procs(procs):
        return {proc['name']:hcsp.Procedure(proc['name'], proc['text'])
                for proc in procs}


infos = [simulator.SimInfo(info['name'], info['text'], outputs=info['outputs'],
                           procedures=convert_procs(info['procedures']))
         for info in hcsp_info if 'parallel' not in info]
try:
    random.seed(0)
    clock = time.perf_counter()
    res = simulator.exec_parallel(
        infos, num_steps=num_steps, num_show=num_show, show_interval=show_interval,
        start_event=start_event)
    print("Time:", time.perf_counter() - clock)

    # # export the time series data of Plant to files
    # directory = "/Users/BEAR/Projects/mars/Examples/AADL/CCS/TCS/data/"
    # s_file = open(directory + 'HCSP_s_3bus_7ms_p1s_noise_con8', 'w')
    # v_file = open(directory + 'HCSP_v_3bus_7ms_p1s_noise_con8', 'w')
    # for time_states in res['time_series']['vehicle_imp']:
    #     if 's' in time_states['state']:
    #         s_file.write(str(time_states['time']) + ',')
    #         s_file.write(str(time_states['state']['s']) + '\n')
    #     if 'v' in time_states['state']:
    #         v_file.write(str(time_states['time']) + ',')
    #         v_file.write(str(time_states['state']['v']) + '\n')
    # s_file.close()
    # v_file.close()
    #
    # obs_file = open(directory + 'HCSP_obs_3bus_7ms_p1s_noise_con8', 'w')
    # for time_states in res['time_series']['emerg_imp']:
    #     if 'obs_pos' in time_states['state']:
    #         obs_file.write(str(time_states['time']) + ',')
    #         obs_file.write(str(time_states['state']['obs_pos']) + '\n')
    # obs_file.close()
    #
    # print("DONE!")

except simulator.SimulatorException as e:
    print(e.error_msg)



"""Translation from SF Chart to Isabelle."""

from ss2hcsp.hcsp import expr
from ss2hcsp.hcsp import hcsp
from ss2hcsp.sf.sf_state import AND_State, Junction, OR_State, GraphicalFunction
from ss2hcsp.sf.sf_convert import SFConvert
from ss2hcsp.matlab import function
from ss2hcsp.matlab import convert


def translate_expr(e):
    if e is None:
        # None in condition means true
        return "Bc True"
    elif isinstance(e, function.AConst):
        return "N %s" % e.value
    elif isinstance(e, function.Var):
        return "V ''%s''" % e.name
    elif isinstance(e, function.OpExpr):
        if e.op_name == "+":
            return "Plus (%s) (%s)" % (translate_expr(e.exprs[0]), translate_expr(e.exprs[1]))
        elif e.op_name == "-" and len(e.exprs) == 1 and isinstance(e.exprs[0], function.AConst):
            return "N (%s)" % (-e.exprs[0].value)
        elif e.op_name == "-":
            return "Minus (%s) (%s)" % (translate_expr(e.exprs[0]), translate_expr(e.exprs[1]))
        elif e.op_name == '*':
            return "Multiply (%s) (%s)" % (translate_expr(e.exprs[0]), translate_expr(e.exprs[1]))
        elif e.op_name == '/':
            return "Divide (%s) (%s)" % (translate_expr(e.exprs[0]), translate_expr(e.exprs[1]))
        else:
            print(e.op_name)
            raise NotImplementedError
    elif isinstance(e, function.BConst):
        if e.value == True:
            return "Bc True"
        else:
            return "Bc False"
    elif isinstance(e, function.RelExpr):
        if e.op == ">":
            return "(%s) [>] (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op == ">=":
            return "(%s) [>=] (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op == "<":
            return "(%s) [<] (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op == "<=":
            return "(%s) [<=] (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op == "==":
            return "(%s) [==] (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op == "!=":
            return "(%s) [!=] (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        else:
            print(e)
            raise NotImplementedError
    elif isinstance(e, function.LogicExpr):
        if e.op_name == '&&':
            return "(%s) && (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op_name == '||':
            return "(%s) || (%s)" % (translate_expr(e.expr1), translate_expr(e.expr2))
        elif e.op_name == '~':
            return "Not (%s)" % (translate_expr(e.expr1))
        else:
            raise NotImplementedError
    elif isinstance(e, function.FunExpr):
        ret_str = ''
        str_end = ''
        for expr in e.exprs:
            ret_str += '([[%s, ' %(translate_expr(expr))
            str_end += ']])'
        ret_str += ('No_Expr' + str_end)
        return ret_str
    elif isinstance(e, function.DirectName):
        return "V ''%s''" % str(e).split('.')[0]
    else:
        print(e)
        print(type(e))
        raise NotImplementedError

def translate_action(c, diagram, is_tran_act=False):
    if isinstance(c, function.Skip):
        return "SKIP"
    elif isinstance(c, function.Assign):
        #print(c.expr)
        if isinstance(c.expr, function.FunExpr):
            ret_str = ''
            str_end = ''
            if isinstance(c.lname, function.Var):
                ret_str += '([[V \'\'%s\'\', ' %(str(c.lname))
                str_end += ']])'
            else:
                for arg in c.lname.args:
                    ret_str += '([[V \'\'%s\'\', ' %(arg)
                    str_end += ']])'
            ret_str += 'No_Expr'
            ret_str += (str_end + "::= ")
            procedures = dict()
            for fun in diagram.diagram.funs:
                procedures[fun.name] = fun
            #print(procedures)
            f = "\'\'" + c.expr.fun_name + "\'\'"
            if isinstance(procedures[c.expr.fun_name], function.Function):
                ret_str += "\'\'" + c.expr.fun_name + "\'\'" + '<'
                ret_str += translate_expr(c.expr)
                ret_str += '>'
            elif isinstance(procedures[c.expr.fun_name], GraphicalFunction):
                ret_str += "\'\'" + c.expr.fun_name + "\'\'" + '<<'
                ret_str += translate_expr(c.expr)
                ret_str += '>>'
            return ret_str
        else:
            if str(c.lname).split('.')[-1] == "data":
                vname = str(c.lname).split('.')[0]
            else:
                vname = c.lname
            return "\'\'%s\'\' ::= %s" % (vname, translate_expr(c.expr))
    elif isinstance(c, function.Sequence):
        return "(%s); (%s)" %(translate_action(c.cmd1, diagram, is_tran_act), translate_action(c.cmd2, diagram, is_tran_act))
    elif isinstance(c, function.FunctionCall):
        if c.func_name == "send":
            args = c.args
            if len(args) == 1 and len(str(args[0]).split('.')) == 1:
                if str(c.args[0]).startswith('M'):
                    return "send3 ''%s'' " % c.args[0]
                else:
                    if is_tran_act:
                        return "send1 ''%s'' True" % c.args[0]
                    else:
                        return "send1 ''%s'' False" % c.args[0]
            #The following two conditions are directed send
            elif len(args) == 2:
                event,direct_name=args[0],args[1]
                direct_name = '\'\', \'\''.join(str(direct_name).split('.'))
                if is_tran_act:
                    return "send2 ''%s'' True [''%s''] " % (event,direct_name)
                else:
                    return "send2 ''%s'' False [''%s''] " % (event,direct_name)
            elif len(args) == 1 and len(str(args[0]).split('.')) > 1:
                tmp = args.split('.')
                event = tmp[-1]
                direct_name = '\'\', \'\''.join(tmp[:-1])
                if is_tran_act:
                    return "send2 ''%s'' True [''%s''] " % (event,direct_name)
                else:
                    return "send2 ''%s'' False [''%s''] " % (event,direct_name)
            else:
                raise NotImplementedError
        elif c.func_name == 'fprintf':
            args = list(c.args)
            if len(args) == 1:
                try:
                    return "print1 ''%s'' " % args[0].value
                except:
                    return "print1 ''test'' "
            else:
                del(args[0])
                expr = ""
                for i in range(0, len(args)):
                    expr += "print2 (V ''" + str(args[i]) + "'')"
                    if i != (len(args)-1):
                        expr += "; "
                return expr
    #matlab functions and graphical fucntions
        else:
            procedures = dict()
            for fun in diagram.diagram.funs:
                procedures[fun.name] = fun
            #print(procedures)
            f = "\'\'" + c.func_name + "\'\'"
            if isinstance(procedures[c.func_name], function.Function):
                ret_str = "No_Expr ::= %s<" %(f)
                str_end = ''
                #print(c.args[0])
                if len(c.args) == 1:
                    return "print1 ''%s'' " % c.args[0].value
                else:
                    for para in c.args:
                        ret_str += '([[%s, ' %(translate_expr(para))
                        str_end += ']])'
                    ret_str += ('No_Expr' + str_end)
                    ret_str += '>'
                    return ret_str
            elif isinstance(procedures[c.func_name], GraphicalFunction):
                ret_str = "No_Expr ::= %s<<" %(f)
                str_end = ''
                for para in c.args:
                    ret_str += '([[%s, ' %(translate_expr(para))
                    str_end += ']])'
                ret_str += ('No_Expr' + str_end)
                ret_str += '>>'
                return ret_str
                
            else:
                raise NotImplementedError
    elif isinstance(c, function.RaiseEvent):
        if len(str(c.event).split('.')) == 1:
            if is_tran_act:
                return "send1 ''%s'' True" % c.event
            else:
                return "send1 ''%s'' False" % c.event
        #The following two conditions are directed send
        else:
            tmp = str(c.event).split('.')
            event = tmp[-1]
            for ssid, state in diagram.all_states.items():
                if state.name == tmp[-2]:
                    direct_name = get_state_path(state)
            if is_tran_act:
                return "send2 ''%s'' True %s " % (event,direct_name)
            else:
                return "send2 ''%s'' False %s " % (event,direct_name)
    else:
        print(type(c))
        return "SKIP"
        #raise NotImplementedError

def translate_actions(cs, diagram, is_tran_act=False):
    actions_str = ''
    for i in range(len(cs)):
        actions_str += translate_action(cs[i], diagram, is_tran_act=is_tran_act)
        if i != len(cs) -1 :
            actions_list = list(actions_str)
            actions_list.insert(0,'(')
            actions_str = ''.join(actions_list)
            actions_str += ');'
    return actions_str

def translate_event(event):
    if event is None:
        return "S []"
    elif isinstance(event, function.BroadcastEvent):
        if event.name.startswith('M'):
            return "M \'\'%s\'\'" % event.name
        else:
            return "S [\'\'%s\'\']" % event.name
    elif isinstance(event, function.TemporalEvent):
        temporal_name = event.temp_op.title()
        now_event = str(event.event)
        if now_event == 'sec':
            now_event = 'Sec'
        return "T1 (%s (\'\'%s\'\') (%s))" % (temporal_name, now_event, translate_expr(event.expr))
    else:
        print(type(event))
        raise NotImplementedError

def get_state_path(state):
    if state.name != '':
        path_list = [state.name]
    else:
        path_list = [str(state.ssid)]
    while state.father:
        if state.father.name != '':
            path_list.append(state.father.name)
        else:
            path_list.append(str(state.father.ssid))
        state = state.father
    path = '['
    for i in range(len(path_list)-2, -1, -1):
        path += ('\'\'' + path_list[i] + '\'\'')
        if i != 0:
            path += ', '
    path += ']'
    return path

def translate_state_name(ssid, diagram, is_GraphicalJunction = False):
    if is_GraphicalJunction:
        if ssid is None:
            return "NONE"
        else:
            return "J [\'\'%s\'\']" %(str(ssid))
    if ssid is None:
        return "NONE"
    
    state = diagram.all_states[ssid]
    path = get_state_path(state)
    #print(type(state))
    #print(state_path)
    if isinstance(state, (OR_State, AND_State)):
        return "P %s" % path
    elif isinstance(state, Junction):
        return "J %s" % path
    else:
        raise NotImplementedError

def translate_trans(tran, diagram, is_GraphicalJunction = False):
    # Obtain source and destination states
    src_str = translate_state_name(tran.src, diagram, is_GraphicalJunction=is_GraphicalJunction)
    dst_str = translate_state_name(tran.dst, diagram, is_GraphicalJunction=is_GraphicalJunction)

    # Obtain event, condition, condition action and transition action
    label = tran.label
    event_str = translate_event(label.event)
    cond_str = translate_expr(label.cond)
    if isinstance(label.cond, function.RelExpr) and isinstance(label.cond.expr1, function.FunExpr):
        cond_str = "(V '''') " + cond_str.split(') ')[1]

    #print(label.tran_act)
    cond_act_str = translate_action(label.cond_act, is_tran_act=False, diagram=diagram)
    tran_act_str = translate_action(label.tran_act, is_tran_act=True, diagram=diagram)

    return "Trans (%s) (%s) (%s) (%s) (%s) (%s)" % \
        (src_str, event_str, cond_str, cond_act_str, tran_act_str, dst_str)

def translate_state(ssid, diagram):
    state = diagram.all_states[ssid]
    if isinstance(state, OR_State):
        en_str = translate_actions(state.en, diagram, is_tran_act=False)
        if en_str == '':
            en_str = 'SKIP'
        du_str = translate_actions(state.du, diagram, is_tran_act=False)
        if du_str == '':
            du_str = 'SKIP'
        ex_str = translate_actions(state.ex, diagram, is_tran_act=False)
        if ex_str == '':
            ex_str = 'SKIP'

        state_path = get_state_path(state)

        if state.inner_trans:
            in_str = ", ".join(translate_trans(tran, diagram) for tran in state.inner_trans)
        else:
            in_str = ""
        
        if state.out_trans:
            out_str = ", ".join(translate_trans(tran, diagram) for tran in state.out_trans)
        else:
            out_str = ""
        for tran in state.out_trans:
            label = tran.label
            tmp_str = ""
            if isinstance(label.cond, function.RelExpr) and isinstance(label.cond.expr1, function.FunExpr):
                procedures = dict()
                for fun in diagram.diagram.funs:
                    procedures[fun.name] = fun
                #print(procedures)
                c = label.cond.expr1
                f = "\'\'" + c.fun_name + "\'\'"
                tmp_str = "((([[V '''', No_Expr]])::="
                if isinstance(procedures[c.fun_name], function.Function):
                    tmp_str += "\'\'" + c.fun_name + "\'\'" + '<'
                    tmp_str += translate_expr(c)
                    tmp_str += '>))'
                elif isinstance(procedures[c.fun_name], GraphicalFunction):
                    tmp_str += "\'\'" + c.fun_name + "\'\'" + '<<'
                    tmp_str += translate_expr(c)
                    tmp_str += '>>))'
                en_list = list(en_str)
                en_list.insert(0,'(')
                en_str = "".join(en_list)
                en_str += "); "
                en_str += tmp_str

        if not state.children:
            comp_str = "No_Composition"
        else:
            comp_str = translate_composition(ssid, diagram)
            if comp_str == '':
                comp_str = "No_Composition"

        return "State %s\n  (%s)\n  (%s)\n  (%s)\n  [%s]\n  [%s]\n  (%s)" % \
            (state_path, en_str, du_str, ex_str, in_str, out_str, comp_str)
    elif isinstance(state, AND_State):
        en_str = translate_actions(state.en, diagram, is_tran_act=False)
        if en_str == '':
            en_str = 'SKIP'
        du_str = translate_actions(state.du, diagram, is_tran_act=False)
        if du_str == '':
            du_str = 'SKIP'
        ex_str = translate_actions(state.ex, diagram, is_tran_act=False)
        if ex_str == '':
            ex_str = 'SKIP'

        state_path = get_state_path(state)

        if state.inner_trans:
            in_str = ", ".join(translate_trans(tran, diagram) for tran in state.inner_trans)
        else:
            in_str = ""

        out_str = ""

        if not state.children:
            comp_str = "No_Composition"
        else:
            comp_str = translate_composition(ssid, diagram)
            if comp_str == '':
                comp_str = "No_Composition"
        
        return "State %s\n  (%s)\n  (%s)\n  (%s)\n  [%s]\n  [%s]\n  (%s)" % \
            (state_path, en_str, du_str, ex_str, in_str, out_str, comp_str)
    else:
        return ""

def translate_composition(ssid, diagram):
    state = diagram.all_states[ssid]
    if isinstance(state.children[0], OR_State):
        
        default_str = '['
        for i in range(len(state.children)):
            default_trans = state.children[i].default_tran
            if not default_trans:
                continue
            else:
                trans_str = translate_trans(default_trans, diagram)
                default_str += trans_str + ','
        if len(default_str) > 1:
            default_str = default_str[:-1] + ']'
        else:
            default_str = '[]'

        history_junc_str = 'False'
        if isinstance(state, OR_State) and state.has_history_junc == True:
            history_junc_str = 'True'
        
        name = state.get_state_whole_name()
        func_str = 'f_' + name
        return "Or (%s)\n (%s) (%s)" %(default_str, history_junc_str, func_str)
    elif isinstance(state.children[0], AND_State):
        children_list = list()
        for i in range(len(state.children)):
            children_list.append(state.children[i].name)
        tmp_str = '\'\', \'\''.join(children_list)
        children_str = '\'\'' + tmp_str + '\'\''

        name = state.get_state_whole_name()
        func_str = 'f_' + name
        return "And [%s]\n (%s)" %(children_str, func_str)

    else:
        return ""

#translate functions for Or/And Composition to get its children
def translate_composition_function(ssid, diagram):
    state = diagram.all_states[ssid]
    name = 'f_' + state.get_state_whole_name()
    fun_str = 'definition %s :: string2state where \n' %name
    def_str = ('%s_def' % name)
    fun_str += '\" %s = \n(λstr. ' %name
    for child in state.children:
        if not isinstance(child, Junction):
            child_name = child.get_state_whole_name()
            fun_str += 'if str = \'\'%s\'\' then %s else \n' %(child.name, child_name)
    fun_str += 'No_State )\"'
    return fun_str, def_str

#translate functions for junction to get its transition list
def translate_junction_function(diagram):
    fun_str = 'definition g :: juncs where \n'
    fun_str += '\" g = (λ str. '
    for ssid, state in diagram.all_states.items():
        if isinstance(state, Junction):
            junc_path = get_state_path(state)
            translist_str = '['
            for tran in state.out_trans:
                translist_str += translate_trans(tran, diagram)
                translist_str += ',\n'
            if translist_str[-1] == '\n':
                translist_str = translist_str[:-2] + ']'
            else:
                translist_str += ']'
            if translist_str != "[]":
                fun_str += 'if str = %s then %s else \n' %(junc_path, translist_str)
    for fun in diagram.diagram.funs:
        if not isinstance(fun, GraphicalFunction):
            continue
        for junction in fun.junctions:
            junc_path = get_state_path(junction)
            translist_str = '['
            for tran in junction.out_trans:
                translist_str += translate_trans(tran, diagram, True)
                translist_str += ',\n'
            if translist_str[-1] == '\n':
                translist_str = translist_str[:-2] + ']'
            else:
                translist_str += ']'
            fun_str += 'if str = %s then %s else \n' %(junc_path, translist_str) 
    fun_str += '[])\"'
    return fun_str

#转化顺序有要求，一定得先从最低一层转换，否则会报错
def dfs_search_chart(ssid, diagram, str, def_list):
    state = diagram.all_states[ssid]
    if isinstance(state, Junction):
        return str, def_list
    if not state.children:
        name = state.get_state_whole_name()
        str += 'definition %s :: state where \" %s = ' %(name, name)
        state_str = translate_state(ssid, diagram)
        str += state_str
        str += '\"\n\n'
        def_list.append('%s_def' %name)
    else:
        for child in state.children:
            child_ssid = child.ssid
            str, def_list = dfs_search_chart(child_ssid, diagram, str, def_list)
        fun_str, def_str = translate_composition_function(ssid, diagram)
        str += fun_str
        str += '\n\n'
        def_list.append(def_str)
        if not ssid == diagram.diagram.ssid:
            name = state.get_state_whole_name()
            str += 'definition %s :: state where \" %s = ' %(name, name)
            state_str = translate_state(ssid, diagram)
            str += state_str
            str += '\"\n\n'
            def_list.append('%s_def' %name)
        else:
            str += 'definition Root :: comp where \" Root = '
            comp_str = translate_composition(ssid, diagram)
            str += comp_str
            str += '\"\n\n'
            def_list.append('Root_def')
    return str, def_list


def translate_chart_info(chart):
    chart_str = 'definition I :: ctxt where \n'
    chart_str += '\"I str = ('
    str1 = '(Info False [] [])'
    for ssid, state in chart.all_states.items():
        if isinstance(state, Junction):
            continue
        state_path = get_state_path(state)
        if not state.children:
            chart_str += 'if str = %s then %s else\n' %(state_path,str1)
        elif isinstance(state.children[0], AND_State):
            chart_str += 'if str = %s then %s else\n' %(state_path,str1)
        elif isinstance(state.children[0], OR_State):
            chart_str += 'if str = %s then %s else\n' %(state_path,str1)
        else:
            continue
    
    chart_str += str1 + ')\"\n'
    return chart_str

def translate_fe_info(chart):
    fe_str = 'fe str = ('
    for fun in chart.diagram.funs:
        if not isinstance(fun, function.Function):
            continue
        fun_name = "\'\'" + fun.name + "\'\'"
        fenv_str = "(" + translate_action(fun.cmd, chart) + ',\n'
        tmp_str = ''
        str_end = ''
        for param in fun.params:
            tmp_str += '([[V \'\'%s\'\', ' %(param)
            str_end += ']])'
        tmp_str += ('No_Expr' + str_end)
        fenv_str += (tmp_str + ',\n')
        tmp_str = ''
        str_end = ''
        if fun.return_var is not None:
            if isinstance(fun.return_var, str):
                tmp_str += '([[V \'\'%s\'\', ' %(fun.return_var)
                str_end += ']])'
                tmp_str += ('No_Expr' + str_end)
            else:
                for var in fun.return_var:
                    tmp_str += '([[V \'\'%s\'\', ' %(var)
                    str_end += ']])'
                tmp_str += ('No_Expr' + str_end)
            fenv_str += (tmp_str + ')')
        else:
            fenv_str += ('No_Expr' + ')')
        fe_str += 'if str = %s then %s else \n' %(fun_name, fenv_str)
    fe_str += (("(SKIP, No_Expr, No_Expr)) \"") + '\n')
    #print(fe_str)
    return fe_str
            #print(procedures)
    
def translate_ge_info(chart):
    ge_str = 'ge str = ('
    for fun in chart.diagram.funs:
        if not isinstance(fun, GraphicalFunction):
            continue
        fun_name = "\'\'" + fun.name + "\'\'"
        genv_str = "((" + translate_trans(fun.default_tran, chart, True) + '),\n'
        tmp_str = ''
        str_end = ''
        for param in fun.params:
            tmp_str += '([[V \'\'%s\'\', ' %(param)
            str_end += ']])'
        tmp_str += ('No_Expr' + str_end)
        genv_str += (tmp_str + ',\n')
        tmp_str = ''
        str_end = ''
        if fun.return_var is not None:
            if isinstance(fun.return_var, str):
                tmp_str += '([[V \'\'%s\'\', ' %(fun.return_var)
                str_end += ']])'
                tmp_str += ('No_Expr' + str_end)
            else:
                for var in fun.return_var:
                    tmp_str += '([[V \'\'%s\'\', ' %(var)
                    str_end += ']])'
                tmp_str += ('No_Expr' + str_end)
            genv_str += (tmp_str + ')')
        else:
            genv_str += ('No_Expr' + ')')
        ge_str += 'if str = %s then %s else \n' %(fun_name, genv_str)
    ge_str += (("((Trans NONE (S []) (Bc True) SKIP SKIP NONE), No_Expr, No_Expr)) \"") + '\n')
    return ge_str
        




from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, RelExpr, AConst, FunExpr, OpExpr, IfExpr
from ss2hcsp.hcsp import hcsp as hp


class Sqrt(SL_Block):
    """Compute the square root"""
    def __init__(self, name, operator, st=-1):
        super(Sqrt, self).__init__("sqrt", name, 1, 1, st)

        assert operator == 'signedSqrt'
        self.operator = operator

    def get_expr(self):
        in_var = AVar(self.dest_lines[0].name)
        return IfExpr(
            RelExpr(">=", in_var, AConst(0)),
            FunExpr("sqrt", [in_var]),
            OpExpr("-", FunExpr("sqrt", [OpExpr("-", in_var)])))

    def __str__(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, expr, str(self.st))

    def __repr__(self):
        return "Sqrt(%s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.operator, self.st, str(self.dest_lines), str(self.src_lines))

    def get_output_hp(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, expr)

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}

    def get_var_map(self):
        in_var = AVar(self.dest_lines[0].name)
        if self.operator == "signedSqrt":
            cond0 = RelExpr("<", in_var, AConst(0))
            expr0 = OpExpr("-", FunExpr("sqrt", [FunExpr("abs", [in_var])]))
            cond1 = cond0.neg()
            expr1 = FunExpr("sqrt", [in_var])
            out_var = self.src_lines[0][0].name
            return {out_var: [(cond0, expr0), (cond1, expr1)]}



"""Unittest for simulation of HCSP."""

import unittest
import math
import pprint

from ss2hcsp.hcsp.hcsp import Channel, Procedure, Function, Skip
from ss2hcsp.hcsp import simulator
from ss2hcsp.hcsp import parser


def run_test(self, infos, num_events, trace, *, io_filter=None,
             print_state=False, print_res=False, warning=None):
    """Test function for HCSP processes.

    infos : List[str, Tuple[Dict[str, HCSP], str] -
        Input HCSP processes. Each process can either be a single HCSP program,
        or a list of procedure specifications followed by the HCSP program.
        Each procedure is specified by a pair of name and HCSP program.

    num_events : int - number of communication or waiting events to simulate.
    trace : List[str] - expected output trace.
    io_filter : str -> bool - which IO events to remain in the trace.
    print_time_series : bool - whether to show time series.
    print_state : bool - whether to show final state.
    warning : [None, str] - if not None, execution must raise a warning whose
        text equals the given string.

    """
    # Process the input HCSP processes, converting them into SimInfo objects
    sim_infos = []
    if isinstance(infos, list):
        for i in range(len(infos)):
            if isinstance(infos[i], str):
                # Single HCSP program
                sim_infos.append(simulator.SimInfo('P' + str(i), infos[i]))
            else:
                # HCSP program with procedure and function specifications
                hp = None
                procedures = dict()
                functions = dict()
                for spec in infos[i]:
                    if isinstance(spec, str):
                        hp = spec
                    elif isinstance(spec, Procedure):
                        procedures[spec.name] = spec
                    elif isinstance(spec, Function):
                        functions[spec.name] = spec
                    else:
                        raise AssertionError("run_test: unknown info")
                sim_infos.append(simulator.SimInfo('P' + str(i), hp, procedures=procedures,
                                 functions=functions))

    elif isinstance(infos, dict):
        procedures = dict()
        for name, (procs, hp) in infos.items():
            
            for proc_name, proc_hp in procs.items():
                procedures[proc_name] = Procedure(proc_name, proc_hp)
        for name, (procs, hp) in infos.items():
            
            # for proc_name, proc_hp in procs.items():
            #     procedures[proc_name] = Procedure(proc_name, proc_hp)
            sim_infos.append(simulator.SimInfo(name, hp, procedures=procedures))
    else:
        raise TypeError

    # Perform the simulation
    res = simulator.exec_parallel(sim_infos, num_io_events=num_events)

    if io_filter is None:
        io_filter = lambda s: True

    # Extract and compare trace of events
    res_trace = [event['str'] for event in res['trace']
                 if event['str'] not in ('start', 'step') and
                    (event['type'] != 'comm' or io_filter(event['ch_name']))]

    if print_res:
        print(res_trace)

    self.assertEqual(res_trace, trace)

    # Optional: print state
    if print_state:
        for info in infos:
            print(info.state)

    # If warning string is set, expect a warning to be raised
    if warning is not None:
        self.assertTrue('warning' in res)
        self.assertEqual(res['warning'], warning)


class SimulatorTest(unittest.TestCase):
    def testEvalExpr(self):
        test_data = [
            ("x + 2", {"x": 1}, 3),
            ("t % 3", {"t": 7}, 1),
        ]

        for expr, state, res in test_data:
            expr = parser.expr_parser.parse(expr)
            info = simulator.SimInfo("P1", Skip(), state=state)
            self.assertEqual(info.eval_expr(expr), res)

    def testEvalExpr(self):
        test_data = [
            ("x >= y", {"x": 3, "y": 2}, True),
            ("x >= y", {"x": 2, "y": 3}, False),
            ("x > y || x < y", {"x": 2, "y": 2}, False),
            ("x == 2 -> y == 3", {"x": 2, "y": 3}, True),
            ("x == 2 -> y == 3", {"x": 2, "y": 4}, False),
            ("x == 2 -> y == 3", {"x": 3, "y": 4}, True),
        ]

        for expr, state, res in test_data:
            expr = parser.expr_parser.parse(expr)
            info = simulator.SimInfo("P1", Skip(), state=state)
            self.assertEqual(info.eval_expr(expr), res)

    def testExecStep(self):
        test_data = [
            ("skip;", (), {}, None, {}),
            ("x := 2;", (), {}, None, {"x": 2}),
            ("x := x + 1;", (), {"x": 2}, None, {"x": 3}),
            ("x := 2; x := x + 1;", (0,), {}, (1,), {"x": 2}),
            ("skip; x := x + 1;", (0,), {"x": 2}, (1,), {"x": 2}),
            ("{ x := x + 1; }*", (0,), {"x": 2}, (0,), {"x": 3}),
            ("{ x := 2; x := x + 1; }*", (0, 0), {"x": 1}, (0, 1), {"x": 2}),
            ("{ x := 2; x := x + 1; }*", (0, 1), {"x": 2}, (0, 0), {"x": 3}),
            ("{ {x_dot = 1 & true} |> [] (p2c!x --> skip;) c2p?x; }*", (0, 0, 0), {}, (0, 1), {}),
            ("if (x == 0) { x := 2; } x := 3;", (0,), {"x": 0}, (0, 0), {"x": 0}),
            ("if (x == 0) { x := 2; } x := 3;", (0,), {"x": 1}, (1,), {"x": 1}),
            ("if (x == 0) { x := 2; }", (), {"x": 0}, (0,), {"x": 0}),
            ("if (x == 0) { x := 2; }", (), {"x": 1}, None, {"x": 1}),
            ("if (x == 0) { x := 1; } else { x := 0; }", (), {"x": 0}, (0,), {"x": 0}),
            ("if (x == 0) { x := 1; } else { x := 0; }", (), {"x": 1}, (1,), {"x": 1}),
            ("if (x == 0) { x := 1; } else { x := 0; }", (0,), {"x": 0}, None, {"x": 1}),
            ("if (x == 0) { x := 1; } else { x := 0; }", (1,), {"x": 1}, None, {"x": 0}),
            ("if (x == 0) { x := 1; skip; } else { skip; }", (0, 0), {"x": 0}, (0, 1), {"x": 1}),
        ]

        for cmd, pos, state, pos2, state2 in test_data:
            info = simulator.SimInfo('P0', cmd, pos=pos, state=state)
            info.exec_step([])
            self.assertEqual(info.reason, None)
            self.assertEqual(info.callstack.top_pos(), pos2)
            self.assertEqual(info.state, state2)

    def testExecStep2(self):
        test_data = [
            ("c?x;", (), {}, {"comm": [(Channel("c"), "?")]}),
            ("c!x;", (), {}, {"comm": [(Channel("c"), "!")]}),
            ("wait(3);", (0,), {}, {"delay": 3}),
            ("wait(3);", (1,), {}, {"delay": 2}),
            ("{x_dot = 1 & true} |> [](c1?x --> skip;, c2!y --> skip;)", (), {},
             {"comm": [(Channel("c1"), "?"), (Channel("c2"), "!")]}),
            ("x := 1; wait(3);", (1, 0), {}, {"delay": 3}),
            ("{x := 1; wait(3);}*", (0, 1, 0), {}, {"delay": 3}),
            ("{x := 1; wait(3);}*", (0, 1, 1), {}, {"delay": 2}),
            ("{x_dot = 1 & x < 1} |> [](c1?x --> skip;, c2!y --> skip;)", (), {"x": 0},
             {"comm": [(Channel("c1"), "?"), (Channel("c2"), "!")], "delay": 1.0}),
            ("{x_dot = 1 & x < 1} |> [](c1?x --> skip;, c2!y --> skip;)", (), {"x": 0.5},
             {"comm": [(Channel("c1"), "?"), (Channel("c2"), "!")], "delay": 0.5}),
        ]

        for cmd, pos, state, reason in test_data:
            info = simulator.SimInfo('P0', cmd, pos=pos, state=state)
            info.exec_step([])
            self.assertEqual(info.reason, reason)
            self.assertEqual(info.callstack.top_pos(), pos)
            self.assertEqual(info.state, state)

    def testExecProcess(self):
        test_data = [
            ("skip;", (), {}, None, {}, "end"),
            ("x := 2;", (), {}, None, {"x": 2}, "end"),
            ("x := 2; x := x + 1;", (0,), {}, None, {"x": 3}, "end"),
            ("x := x + 1; c!x;", (0,), {"x": 2}, (1,), {"x": 3}, {"comm": [(Channel("c"), "!")]}),
            ("wait(3);", (0,), {}, (0,), {}, {"delay": 3}),
            ("{ x := x + 1; wait(3); }*", (0, 0), {"x": 2}, (0, 1, 0), {"x": 3}, {"delay": 3}),
            ("if (x > 0) { x := 1; } if (x < 0) { x := -1; }", (0,), {"x": 0}, None, {"x": 0}, "end"),
            ("if (x > 0) { x := 1; } if (x < 0) { x := -1; }", (0,), {"x": 2}, None, {"x": 1}, "end"),
            ("if (x > 0) { x := 1; } if (x < 0) { x := -1; }", (0,), {"x": -2}, None, {"x": -1}, "end"),
        ]

        for cmd, pos, state, pos2, state2, reason in test_data:
            info = simulator.SimInfo('P0', cmd, pos=pos, state=state)
            while info.callstack.top_pos() is not None:
                info.exec_step([])
                if info.reason is not None:
                    break
            if info.callstack.top_pos() is None:
                info.reason = "end"
            self.assertEqual(info.reason, reason)
            self.assertEqual(info.callstack.top_pos(), pos2)
            self.assertEqual(info.state, state2)

    def testExecInputComm(self):
        test_data = [
            ("c?x;", (), {}, "c", 2, None, {"x": 2}),
            ("c?x; x := x + 1;", (0,), {}, "c", 2, (1,), {"x": 2}),
            ("c?x; x := x + 1; y := 2;", (0,), {}, "c", 2, (1,), {"x": 2}),
            ("{x_dot = 1 & true} |> [](c?x --> x := x + 1;)", (), {}, "c", 2, (0,), {"x": 2}),
            ("{x_dot = 1 & true} |> [](c?x --> skip;) x := x + 2;", (0,), {}, "c", 2, (0, 0), {"x": 2})
        ]

        for cmd, pos, state, ch_name, val, pos2, state2 in test_data:
            info = simulator.SimInfo('P0', cmd, pos=pos, state=state)
            info.exec_input_comm(Channel(ch_name), val)
            self.assertEqual(info.callstack.top_pos(), pos2)
            self.assertEqual(info.state, state2)

    def testExecOutputComm(self):
        test_data = [
            ("c!x;", (), {"x": 2}, "c", None, 2, {"x": 2}),
            ("c!x+1;", (), {"x": 2}, "c", None, 3, {"x": 2}),
            ("c!x+y; x := 3;", (0,), {"x": 2, "y": 3}, "c", (1,), 5, {"x": 2, "y": 3}),
            ("c!x*y; x := 3; y := 0;", (0,), {"x": 2, "y": 3}, "c", (1,), 6, {"x": 2, "y": 3}),
            ("{x_dot = 1 & true} |> [](c!x --> skip;)", (), {"x": 2}, "c", (0,), 2, {"x": 2}),
            ("{x_dot = 1 & true} |> [](c!x+1 --> skip;) x := x + 1;", (0,), {"x": 2}, "c", (0, 0), 3, {"x": 2}),
        ]

        for cmd, pos, state, ch_name, pos2, val, state2 in test_data:
            info = simulator.SimInfo('P0', cmd, pos=pos, state=state)
            res = info.exec_output_comm(Channel(ch_name))
            self.assertEqual(res, val)
            self.assertEqual(info.callstack.top_pos(), pos2)
            self.assertEqual(info.state, state2)

    def testExecDelay(self):
        test_data = [
            ("wait(3);", (0,), {}, 3, None, {}),
            ("wait(3);", (0,), {}, 2, (2,), {}),
            ("wait(3);", (1,), {}, 1, (2,), {}),
            ("wait(3);", (1,), {}, 2, None, {}),
            ("wait(3); x := x + 1;", (0, 0), {"x": 2}, 3, (1,), {"x": 2}),
            ("wait(3); x := x + 1;", (0, 0), {"x": 2}, 2, (0, 2), {"x": 2}),
            ("wait(3); x := x + 1;", (0, 1), {"x": 2}, 2, (1,), {"x": 2}),
            ("{x_dot = 1 & true} |> [](c!x --> skip;)", (), {"x": 2}, 3, (), {"x": 5}),
            ("{x_dot = 1 & true} |> [](c!x --> skip;) x := x + 1;", (0,), {"x": 2}, 3, (0,), {"x": 5}),
        ]

        for cmd, pos, state, delay, pos2, state2 in test_data:
            info = simulator.SimInfo('P0', cmd, pos=pos, state=state)
            info.exec_step([])  # obtain delay value
            info.exec_delay(delay)
            self.assertEqual(info.callstack.top_pos(), pos2)
            self.assertEqual(info.state, state2)

    def assertAlmostEqualState(self, st1, st2):
        self.assertEqual(set(st1.keys()), set(st2.keys()))
        for k in st1:
            self.assertAlmostEqual(st1[k], st2[k], places=3)

    def testODEComm(self):
        test_data = [
            # Acceleration
            ("{x_dot = v, v_dot = a & true} |> [](c!x --> skip;)",
             {"a": 1, "v": 0, "x": 0}, 3, {"a": 1, "v": 3, "x": 4.5}),

            # Acceleration with floating point numbers
            ("{x_dot = v, v_dot = a & true} |> [](c!x --> skip;)",
             {"a": 1.1, "v": 0, "x": 0}, 3, {"a": 1.1, "v": 3.3, "x": 4.95}),

            # Exponential growth
            ("{x_dot = x & true} |> [](c!x --> skip;)", {"x": 1}, 3, {"x": math.exp(3)}),

            # Circular motion
            ("{x_dot = -1 * y, y_dot = x & true} |> [](c!x --> skip;)",
             {"x": 1, "y": 0}, 3, {"x": math.cos(3), "y": math.sin(3)}),
        ]

        for cmd, state, delay, state2 in test_data:
            info = simulator.SimInfo('P0', cmd, state=state)
            info.exec_step([])  # obtain delay value
            info.exec_delay(delay)
            self.assertAlmostEqualState(info.state, state2)

    def testODEDelay(self):
        test_data = [
            # Acceleration
            ("{x_dot = v, v_dot = a, t_dot = 1 & t < 3}", {"t": 0, "a": 1, "v": 0, "x": 0}, 3.0),

            # Acceleration, upon x reaching a certain value
            ("{x_dot = v, v_dot = a, t_dot = 1 & x < 3}",
             {"t": 0, "a": 1, "v": 0, "x": 0}, math.sqrt(6)),

            # Exponential growth
            ("{x_dot = x, t_dot = 1 & x < 3}", {"t": 0, "x": 1}, math.log(3)),

            # Circular motion
            ("{x_dot = -1 * y, y_dot = x & x > 0}", {"x": 1, "y": 0}, math.pi/2),

            # Some examples of large or infinite delay
            ("{t_dot = 0.1 & t < 6}", {"t": 0}, 60),
            ("{t_dot = 0.01 & t < 60}", {"t": 0}, 100),  # maximum is 100
            ("{x_dot = -1 * y, y_dot = x & x > -2}", {"x": 1, "y": 0}, 100),

            # Conjunction
            ("{x_dot = 1, y_dot = 1 & x < 3 && y < 2}", {"x": 0, "y": 0}, 2.0),
            ("{x_dot = 1, y_dot = -1 & x < 3 && y >= 0}", {"x": 0, "y": 1}, 1.0),
        ]

        for cmd, state, delay in test_data:
            hp = parser.hp_parser.parse(cmd)
            info = simulator.SimInfo("P1", Skip(), state=state)
            res = info.get_ode_delay(hp)
            self.assertAlmostEqual(res, delay, places=5)

    def testExecParallel1(self):
        run_test(self, [
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*",
            "{ wait(2); p2c?x; c2p!x-1; }*",
        ], 6, ["delay 2", "IO p2c 2.0", "IO c2p 1.0", "delay 2", "IO p2c 3.0", "IO c2p 2.0"])

    def testExecParallel2(self):
        run_test(self, [
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*",
            "wait(2); p2c?x; c2p!x-1;",
        ], 6, ["delay 2", "IO p2c 2.0", "IO c2p 1.0", "deadlock"])

    def testExecParallel3(self):
        run_test(self, [
            "x := 0; x := x + 1;",
            "y := 2;",
        ], 6, ['deadlock'])

    def testExecParallel4(self):
        run_test(self, [
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*",
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x; }*",
        ], 6, ['deadlock'])

    def testExecParallel5(self):
        run_test(self, [
            "z := 1; {x?x --> skip; $ z!z --> skip; $ y?y --> skip;}",
            "y := 2; y!y;",
        ], 3, ["IO y 2", "deadlock"])

    def testExecParallel6(self):
        run_test(self, [
            "z := 1; {x?x --> skip; $ z!z --> skip; $ y?y --> skip;}",
            "z?z;",
        ], 3, ["IO z 1", "deadlock"])

    def testExecParallel7(self):
        run_test(self, [
            "x := 1; y := 2; z := 3; wait(3); w := 4;",
            "x := 11; y := 12; wait(2); z := 3;"
        ], 6, ["delay 2", "delay 1", "deadlock"])

    def testExecParallel8(self):
        run_test(self, [
            "{x?x --> x!x+1; $ y?y --> skip;} x!x+2;",
            "x!3; x?x; x?x;"
        ], 3, ["IO x 3", "IO x 4", "IO x 5"])

    def testExecParallel9(self):
        run_test(self, [
            "x := 0; v := 0; a := 1; {x_dot = v, v_dot = a & true} |> [](c!x --> skip;)",
            "wait(3); c?x;"
        ], 3, ["delay 3", "IO c 4.5", "deadlock"])

    def testExecParallel10(self):
        run_test(self, [
            "x := 0; v := 0; a := 1; {x_dot = v, v_dot = a & x < 3} c!x;",
            "c?x;"
        ], 3, ["delay 2.449", "IO c 3.0", "deadlock"])

    def testExecParallel11(self):
        run_test(self, [
            "x := 0; v := 0; a := 1; {x_dot = v, v_dot = a & x < 3} c!x;",
            "x := 0; v := 0; a := 1; {x_dot = v, v_dot = a & x < 5} c!x;",
            "c?x; c?x;"
        ], 5, ["delay 2.449", "IO c 3.0", "delay 0.713", "IO c 5.0", "deadlock"])

    def testExecParallel12(self):
        run_test(self, [
            "x := 0; v := 1; a := -1; { {x_dot = v, v_dot = a & x >= 0} v := -0.8 * v; }*",
        ], 3, ["delay 2.0", "delay 1.6", "delay 1.28"])

    def testExecParallel15(self):
        run_test(self, [
            "x := 1; { {x_dot = x & true} |> [](c!x --> skip;)}*",
            "{ wait(10); c?x; }*"
        ], 100, ['delay 10', 'IO c 22026.814', 'delay 10', 'IO c 485180534.947', 'delay 10', 'overflow'])

    def testExecParallel16(self):
        run_test(self, [
            "EL := []; EL := push(EL, \"a\"); EL := push(EL, \"b\"); EL := pop(EL); x := top(EL); ch!x;",
            "ch?x;"
        ], 2, ['IO ch "a"', "deadlock"])

    def testExecParallel17(self):
        run_test(self, [
            "EL := []; { in?e --> EL := push(EL, e); $ out? --> e := top(EL); outval!e; EL := pop(EL); }*",
            "in!\"a\"; in!\"b\"; out!; outval?e; out!; outval?e;",
        ], 7, ['IO in "a"', 'IO in "b"', 'IO out', 'IO outval "b"', 'IO out', 'IO outval "a"', 'deadlock'])

    def testExecParallel18(self):
        run_test(self, [
            "num := 0; if (num == 0) { E := \"e\"; num := 1; } ch!E;",
            "ch?E;"
        ], 2, ['IO ch "e"', 'deadlock'])

    def testExecParallel9(self):
        run_test(self, [
            "num := 0; {ch_a!0 --> skip; $ ch_b?x --> skip;}*",
            "ch_a?y; ch_b!y;"
        ], 3, ['IO ch_a 0', 'IO ch_b 0', 'deadlock'])

    def testExecParallel20(self):
        run_test(self, [
            "x := 0; y := 0; { {x_dot = 1, y_dot = 1 & true} |> [](cx!x --> cy!y; x := x - 1;, cy!y --> cx!x; y := y - 1;)}*",
            "wait(1); cx?x; cy?y; wait(1); cy?y; cx?x; wait(1); cx?x; cy?y;"
        ], 10, ['delay 1', 'IO cx 1.0', 'IO cy 1.0', 'delay 1', 'IO cy 2.0', 'IO cx 1.0', 'delay 1', 'IO cx 2.0', 'IO cy 2.0', 'deadlock'])

    def testExecParallel21(self):
        run_test(self, [
            "x := 0; {if (x == 0) { x := 1; } else if (x == 1) { x := 2; } else { x := 0; } c!x; }*",
            "{ c?x; }*"
        ], 6, ['IO c 1', 'IO c 2', 'IO c 0', 'IO c 1', 'IO c 2', 'IO c 0'])

    def testExecParallel24(self):
        run_test(self, [
            'x := 0; {x_dot = 1 & x < 1} |> [](c!x --> skip;) {x_dot = 1 & x < 1} |> [](c!x --> skip;) c!x;',
            'wait(0.2); c?x; wait(1.3); c?x;'
        ], 10, ['delay 0.2', 'IO c 0.2', 'delay 0.8', 'delay 0.5', 'IO c 1.0', 'deadlock'])

    def testExecParallel25(self):
        run_test(self, [
            'x := 0; { {x_dot = 1 & x < 1} |> [](c!x --> skip;)}*(x < 1) c!x;',
            'wait(0.2); c?x; wait(1.3); c?x;'
        ], 10, ['delay 0.2', 'IO c 0.2', 'delay 0.8', 'delay 0.5', 'IO c 1.0', 'deadlock'])

    def testExecParallel26(self):
        run_test(self, [
            'x := 0; y := 0; {x_dot = 1, y_dot = 2 & x < 2 && y < 3} c!x; c!y;',
            'c?x; c?x;'
        ], 4, ['delay 1.5', 'IO c 1.5', 'IO c 3.0', 'deadlock'])

    def testExecParallel27(self):
        run_test(self, [
            'xs := []; {ch_new?new; xs := push(xs, new); x := get_max(xs); ch_max!x; xs := pop_max(xs);}*',
            'ch_new![1,2]; ch_max?x; ch_new!3; ch_max?x; ch_new![]; ch_max?x;'
        ], 10, ['IO ch_new [1,2]', 'IO ch_max 2', 'IO ch_new 3', 'IO ch_max 3', 'IO ch_new []', 'IO ch_max 1', 'deadlock'])

    def testExecParallel28(self):
        run_test(self, [
            'b := 10; xs := [b, 1]; a := xs[0]; ch!a; b := xs[1]; ch!b;',
            'ch?a; ch?b;'
        ], 3, ['IO ch 10', 'IO ch 1', 'deadlock'])

    def testExecParallel29(self):
        """Test multiple assignment."""
        run_test(self, [
            'b := 10; xs := [b, 1]; (a, b) := xs; ch!a; ch!b;',
            'ch?a; ch?b;'
        ], 3, ['IO ch 10', 'IO ch 1', 'deadlock'])

    def testExecParallel30(self):
        run_test(self, [
            'x := 0; {x_dot = 0 * x & x <= 0}'
        ], 3, ['delay 100', 'deadlock'])

    def testExecParallel31(self):
        run_test(self, [
            'x := 0; {x_dot = 0 * x & x < 0}'
        ], 3, ['delay 0.0', 'deadlock'])

    def testExecParallel32(self):
        run_test(self, [
            'x := 0; v := 0; {x_dot = v, v_dot = 0 & x <= 0}'
        ], 3, ['delay 100', 'deadlock'])

    def testExecParallel33(self):
        run_test(self, [
            'x := 2; assert(x == 2);'
        ], 2, ['deadlock'])

    def testExecParallel34(self):
        run_test(self, [
            'x := 2; test(x == 3, "x should equal 3");'
        ], 3, ['warning: Test x == 3 failed (x should equal 3)', 'deadlock'],
        warning=(0, "Test x == 3 failed (x should equal 3)"))

    def testExecParallel35(self):
        run_test(self, [
            'x := 2; log("start");'
        ], 1, ['log start', 'deadlock'])

    def testExecParallel36(self):
        run_test(self, [
            'x := 0; { {x_dot = 1 & true} |> [](p2c[0]!x --> skip;) c2p[0]?x;}*',
            '{wait(2); p2c[0]?x; c2p[0]!x-1;}*'
        ], 6, ['delay 2', 'IO p2c[0] 2.0', 'IO c2p[0] 1.0', 'delay 2', 'IO p2c[0] 3.0', 'IO c2p[0] 2.0'])

    def testExecParallel37(self):
        run_test(self, [
            'x := 0; i := 0; {i := i + 1; {x_dot = 1 & true} |> [](p2c[i]!x --> skip;) c2p[i]?x;}*',
            'i := 0; {i := i + 1; wait(2); p2c[i]?x; c2p[i]!x-1;}*'
        ], 6, ['delay 2', 'IO p2c[1] 2.0', 'IO c2p[1] 1.0', 'delay 2', 'IO p2c[2] 3.0', 'IO c2p[2] 2.0'])

    def testExecParallel38(self):
        run_test(self, [
            'pt := {x:1, y:2}; pt.x := pt.y + 1; ch!pt.x; ch!pt;',
            'ch?x; ch?x;'
        ], 3, ['IO ch 3', 'IO ch {x:3,y:2}', 'deadlock'])

    def testExecParallel39(self):
        run_test(self, [
            'pt := [1, 2]; pt[0] := pt[1] + 1; ch!pt[0]; ch!pt;',
            'ch?x; ch?x;'
        ], 3, ['IO ch 3', 'IO ch [3,2]', 'deadlock'])

    def testExecParallel40(self):
        run_test(self, [
            'pt := {x:1, y:2}; pt2 := pt; pt2.y := 3; ch!pt2.y; ch!pt.y;',
            'ch?x; ch?x;'
        ], 3, ['IO ch 3', 'IO ch 2', 'deadlock'])

    def testExecParallel41(self):
        run_test(self, [
            'pt := {x:1, y:2}; ch!pt; ch!pt.y;',
            'ch?pt; pt.y := 3; ch?y;'
        ], 3, ['IO ch {x:1,y:2}', 'IO ch 2', 'deadlock'])

    def testExecParallel42(self):
        run_test(self, [
            'pt := {x:1, y:2}; ch!pt.x; ch!pt.y; ch?z; ch?z;',
            'pt := {x:2, y:3}; ch?pt.x; ch?pt.y; ch!pt.x; ch!pt.y;'
        ], 5, ['IO ch 1', 'IO ch 2', 'IO ch 1', 'IO ch 2', 'deadlock'])

    def testExecParallel43(self):
        run_test(self, [
            "x := 0; {test(x <= 2, \"x is too big\"); {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x;}*",
            "{wait(2); p2c?x; c2p!x-1;}*",
        ], 12, ["delay 2", "IO p2c 2.0", "IO c2p 1.0", "delay 2", "IO p2c 3.0", "IO c2p 2.0",
                "delay 2", "IO p2c 4.0", "IO c2p 3.0", "warning: Test x <= 2 failed (x is too big)", "delay 2", "IO p2c 5.0", "IO c2p 4.0"],
        warning=(6, "Test x <= 2 failed (x is too big)"))

    def testExecParallel44(self):
        run_test(self, [
            "status := [1]; status[0] := {x:1, y:2}; ch!status[0].x; ch!status[0].y;",
            "ch?z; ch?x;"
        ], 3, ['IO ch 1', 'IO ch 2', 'deadlock'])

    def testExecParallel45(self):
        run_test(self, [
            "a := [[1,2],[3,4]]; ch!a[1][0]; ch!a[0][1];",
            "ch?x; ch?x;"
        ], 3, ['IO ch 3', 'IO ch 2', 'deadlock'])

    def testExecParallel46(self):
        run_test(self, [
            "pt := {xs: [1, 2], ys: [3, 4]}; ch!pt.xs[1]; ch!pt.ys[0];",
            "ch?x; ch?x;"
        ], 3, ['IO ch 2', 'IO ch 3', 'deadlock'])

    def testExecParallel47(self):
        run_test(self, [
            "x := 0; {assert(x <= 2, \"x is too big\"); {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x;}*",
            "{wait(2); p2c?x; c2p!x-1;}*",
        ], 12, ["delay 2", "IO p2c 2.0", "IO c2p 1.0", "delay 2", "IO p2c 3.0", "IO c2p 2.0",
                "delay 2", "IO p2c 4.0", "IO c2p 3.0", "error: x is too big"],
        warning=(6, "x is too big"))

    def testExecParallel48(self):
        run_test(self, [
            "{x := 0; {x_dot = 1 & true} |> [](ch[_thread]? --> out!_thread;)}*",
            "ch[0]!; out?x; ch[1]!; out?x;"
        ], 4, ['IO ch[0]', 'IO out 0', 'IO ch[1]', 'IO out 1'])

    def testExecParallel49(self):
        run_test(self, [
            "a := [[0,1],[2,3]]; b := [0,1,2]; a[0][1] := 4; a[1][0] := 5; b[1] := a[1][1]; ch!a; ch!b;",
            "ch?x; ch?y;"
        ], 3, ['IO ch [[0,4],[5,3]]', 'IO ch [0,3,2]', 'deadlock'])

    def testProcedure1(self):
        run_test(self, [
            (Procedure("incr", "x := x + 1;"),
             "x := 0; @incr; @incr; ch!x;"),
            "ch?x;"
        ], 2, ['IO ch 2', 'deadlock'])

    def testProcedure2(self):
        run_test(self, [
            (Procedure("fact", "if (a > 0) { x := a * x; a := a - 1; @fact; } else { skip; }"),
             "x := 1; a := 5; @fact; ch!x;"),
            "ch?x;"
        ], 2, ['IO ch 120', 'deadlock'])

    def testProcedure3(self):
        run_test(self, [
            (Procedure("output", "if (a > 0) { a := a - 1; ch!a; @output; } else { skip; }"),
             "a := 5; @output;"),
            "{ch?x;}*"
        ], 10, ['IO ch 4', 'IO ch 3', 'IO ch 2', 'IO ch 1', 'IO ch 0', 'deadlock'])

    def testProcedure4(self):
        run_test(self, [
            (Procedure("delay", "if (a > 0) { a := a - 1; wait(2); @delay; } else { skip; }"),
             "a := 5; @delay;")
        ], 10, ['delay 2', 'delay 2', 'delay 2', 'delay 2', 'delay 2', 'deadlock'])

    def testProcedure5(self):
        run_test(self, [
            (Procedure("pa", "if (x > 0) { x := x - 1; cha!x; @pb; } else { skip; }"),
             Procedure("pb", "if (x > 0) { x := x - 1; chb!x; @pa; } else { skip; }"),
             "x := 5; @pa;"),
            "{cha?x --> { skip; } $ chb?x --> { skip; }}*"
        ], 10, ['IO cha 4', 'IO chb 3', 'IO cha 2', 'IO chb 1', 'IO cha 0', 'deadlock'])

    def testFunctions1(self):
        run_test(self, [
            (Function("bar", ["a", "b"], "2 * a + 3 * b"),
             "x := 2; y := 3; ch!bar(x, y);"),
            "ch?x;"
        ], 10, ['IO ch 13', 'deadlock'])
    
    def testFunctions2(self):
        run_test(self, [
            (Function("bar", ["a", "b"], "2 * a == b"),
             "x := 2; y := 1; if (bar(y, x)) { ch!x; } else { ch!y; }"),
            "ch?x;"
        ], 10, ['IO ch 2', 'deadlock'])

    def run_test_trace(self, infos, *, num_steps, num_show, ids=None,
                       show_interval=None, start_event=None, print_trace=False):
        for i in range(len(infos)):
            infos[i] = simulator.SimInfo('P' + str(i), infos[i])

        res = simulator.exec_parallel(
            infos, num_steps=num_steps, num_show=num_show,
            show_interval=show_interval, start_event=start_event)
        if print_trace:
            pprint.pprint(res['trace'])
        if ids:
            self.assertEqual(ids, [event['id'] for event in res['trace']])

    def testTrace1(self):
        self.run_test_trace([
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x;}*",
            "{wait(2); p2c?x; c2p!x-1;}*",
        ], num_steps=10, num_show=5, ids=[0,1,2,3,4,5])

    def testTrace2(self):
        self.run_test_trace([
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x;}*",
            "{wait(2); p2c?x; c2p!x-1;}*",
        ], num_steps=10, num_show=5, show_interval=2, ids=[0,2,4,6,8,10])

    def testTrace3(self):
        self.run_test_trace([
            "x := 0; { {x_dot = 1 & true} |> [](p2c!x --> skip;) c2p?x;}*",
            "{wait(2); p2c?x; c2p!x-1;}*",
        ], num_steps=10, num_show=3,
        start_event = {
            'id': 4,
            'infos': {'P0': {'pos': 'p1,0,1', 'state': {'x': 2.0}},
                      'P1': {'pos': 'p0,2', 'state': {'x': 2.0}}},
            'time': 2},
        ids=[5,6,7,8])


if __name__ == "__main__":
    unittest.main()



"""Topological sort"""

class TopologicalSortException(Exception):
    def __init__(self, cycle):
        self.cycle = cycle


def topological_sort(graph):
    """Given a graph stored as adjacency lists, return topological
    order of vertices, or return an error if there is a cycle.
    
    """
    # Form the reverse graph
    rev_graph = dict()
    for v in graph:
        rev_graph[v] = list()
    for v, edges in graph.items():
        for edge in edges:
            rev_graph[edge].append(v)

    # List of roots (no vertices depend on them)
    roots = list()
    for v in graph:
        if len(rev_graph[v]) == 0:
            roots.append(v)

    # Contains topological order
    topo_sort = []

    # Whether a node is explored
    explored = dict()
    for v in graph:
        explored[v] = False

    # Temporary path
    path = []

    def dfs(v):
        for next in graph[v]:
            if next in path:
                # Back-edge, found a loop
                raise TopologicalSortException(path[path.index(next):])
            elif explored[next]:
                # Already explored
                continue
            else:
                # Never explored
                path.append(next)
                explored[next] = True
                dfs(next)
                del path[-1]
        topo_sort.append(v)
    
    for root in roots:
        dfs(root)

    if len(topo_sort) == len(graph):
        return topo_sort
    else:
        for v in graph:
            if not explored[v]:
                dfs(v)  # Must find cycle and raise exception



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.parser import hp_parser


class Integrator(SL_Block):
    """Block for integration."""
    def __init__(self, name, init_value=0):
        super(Integrator, self).__init__("integrator", name, 1, 1, st=0)

        assert isinstance(init_value, (int, float))
        self.init_value = init_value

    def __str__(self):
        in_var = self.dest_lines[0].name
        out_var = self.src_lines[0][0].name
        return "%s: %s_dot = %s  (init = %s)" % (self.name, out_var, in_var, self.init_value)

    def __repr__(self):
        return "Integrator(%s, %s, in = %s, out = %s)" % \
            (self.name, self.init_value, str(self.dest_lines),  str(self.src_lines))



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import true_expr, AConst


class Constant(SL_Block):
    """Block for constant."""
    def __init__(self, name, value):
        super(Constant, self).__init__("constant", name, 1, 0, st=0)

        assert isinstance(value, (int, float))
        self.value = value

    def __str__(self):
        out_var = self.src_lines[0][0].name
        return "%s: %s = %s" % (self.name, out_var, self.value)

    def __repr__(self):
        return "Constant(%s, %s, out = %s)" % (self.name, self.value, str(self.src_lines))

    def get_var_subst(self):
        out_var = self.src_lines[0][0].name
        return {out_var: AConst(self.value)}

    def get_var_map(self):
        out_var = self.src_lines[0][0].name
        expr = AConst(self.value)
        return {out_var: [(true_expr, expr)]}
